# -*- coding: utf-8 -*-
'''
Module for managing IIS SMTP server configuration on Windows servers.
The Windows features 'SMTP-Server' and 'Web-WMI' must be installed.

:depends: wmi

'''

# IIS metabase configuration settings:
#   https://goo.gl/XCt1uO
# IIS logging options:
#   https://goo.gl/RL8ki9
#   https://goo.gl/iwnDow
# MicrosoftIISv2 namespace in Windows 2008r2 and later:
#   http://goo.gl/O4m48T
# Connection and relay IPs in PowerShell:
#   https://goo.gl/aBMZ9K
#   http://goo.gl/MrybFq

# Import python libs
from __future__ import absolute_import
import logging
import re

# Import 3rd-party libs
import salt.ext.six as six

# Import salt libs
from salt.exceptions import SaltInvocationError
import salt.utils

try:
    import wmi
    import salt.utils.winapi
    _HAS_MODULE_DEPENDENCIES = True
except ImportError:
    _HAS_MODULE_DEPENDENCIES = False

_DEFAULT_SERVER = 'SmtpSvc/1'
_WMI_NAMESPACE = 'MicrosoftIISv2'
_LOG = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = 'win_smtp_server'


def __virtual__():
    '''
    Only works on Windows systems.
    '''
    if salt.utils.is_windows() and _HAS_MODULE_DEPENDENCIES:
        return __virtualname__
    return False


def _get_wmi_setting(wmi_class_name, setting, server):
    '''
    Get the value of the setting for the provided class.
    '''
    with salt.utils.winapi.Com():
        try:
            connection = wmi.WMI(namespace=_WMI_NAMESPACE)
            wmi_class = getattr(connection, wmi_class_name)

            objs = wmi_class([setting], Name=server)[0]
            ret = getattr(objs, setting)
        except wmi.x_wmi as error:
            _LOG.error('Encountered WMI error: %s', error.com_error)
        except (AttributeError, IndexError) as error:
            _LOG.error('Error getting %s: %s', wmi_class_name, error)
    return ret


def _set_wmi_setting(wmi_class_name, setting, value, server):
    '''
    Set the value of the setting for the provided class.
    '''
    with salt.utils.winapi.Com():
        try:
            connection = wmi.WMI(namespace=_WMI_NAMESPACE)
            wmi_class = getattr(connection, wmi_class_name)

            objs = wmi_class(Name=server)[0]
        except wmi.x_wmi as error:
            _LOG.error('Encountered WMI error: %s', error.com_error)
        except (AttributeError, IndexError) as error:
            _LOG.error('Error getting %s: %s', wmi_class_name, error)

        try:
            setattr(objs, setting, value)
            return True
        except wmi.x_wmi as error:
            _LOG.error('Encountered WMI error: %s', error.com_error)
        except AttributeError as error:
            _LOG.error('Error setting %s: %s', setting, error)
    return False


def _normalize_server_settings(**settings):
    '''
    Convert setting values that had been improperly converted to a dict back to a string.
    '''
    ret = dict()
    settings = salt.utils.clean_kwargs(**settings)

    for setting in settings:
        if isinstance(settings[setting], dict):
            _LOG.debug('Fixing value: %s', settings[setting])
            value_from_key = next(six.iterkeys(settings[setting]))

            ret[setting] = "{{{0}}}".format(value_from_key)
        else:
            ret[setting] = settings[setting]
    return ret


def get_log_format_types():
    '''
    Get all available log format names and ids.

    :return: A dictionary of the log format names and ids.
    :rtype: dict

    CLI Example:

    .. code-block:: bash

        salt '*' win_smtp_server.get_log_format_types
    '''
    ret = dict()
    prefix = 'logging/'

    with salt.utils.winapi.Com():
        try:
            connection = wmi.WMI(namespace=_WMI_NAMESPACE)
            objs = connection.IISLogModuleSetting()

            # Remove the prefix from the name.
            for obj in objs:
                name = str(obj.Name).replace(prefix, '', 1)
                ret[name] = str(obj.LogModuleId)
        except wmi.x_wmi as error:
            _LOG.error('Encountered WMI error: %s', error.com_error)
        except (AttributeError, IndexError) as error:
            _LOG.error('Error getting IISLogModuleSetting: %s', error)

    if not ret:
        _LOG.error('Unable to get log format types.')
    return ret


def get_servers():
    '''
    Get the SMTP virtual server names.

    :return: A list of the SMTP virtual servers.
    :rtype: list

    CLI Example:

    .. code-block:: bash

        salt '*' win_smtp_server.get_servers
    '''
    ret = list()

    with salt.utils.winapi.Com():
        try:
            connection = wmi.WMI(namespace=_WMI_NAMESPACE)
            objs = connection.IIsSmtpServerSetting()

            for obj in objs:
                ret.append(str(obj.Name))
        except wmi.x_wmi as error:
            _LOG.error('Encountered WMI error: %s', error.com_error)
        except (AttributeError, IndexError) as error:
            _LOG.error('Error getting IIsSmtpServerSetting: %s', error)

    _LOG.debug('Found SMTP servers: %s', ret)
    return ret


def get_server_setting(settings, server=_DEFAULT_SERVER):
    '''
    Get the value of the setting for the SMTP virtual server.

    :param str settings: A list of the setting names.
    :param str server: The SMTP server name.

    :return: A dictionary of the provided settings and their values.
    :rtype: dict

    CLI Example:

    .. code-block:: bash

        salt '*' win_smtp_server.get_server_setting settings="['MaxRecipients']"
    '''
    ret = dict()

    if not settings:
        _LOG.warning('No settings provided.')
        return ret

    with salt.utils.winapi.Com():
        try:
            connection = wmi.WMI(namespace=_WMI_NAMESPACE)
            objs = connection.IIsSmtpServerSetting(settings, Name=server)[0]

            for setting in settings:
                ret[setting] = str(getattr(objs, setting))
        except wmi.x_wmi as error:
            _LOG.error('Encountered WMI error: %s', error.com_error)
        except (AttributeError, IndexError) as error:
            _LOG.error('Error getting IIsSmtpServerSetting: %s', error)
    return ret


def set_server_setting(settings, server=_DEFAULT_SERVER):
    '''
    Set the value of the setting for the SMTP virtual server.

    .. note::

        The setting names are case-sensitive.

    :param str settings: A dictionary of the setting names and their values.
    :param str server: The SMTP server name.

    :return: A boolean representing whether all changes succeeded.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' win_smtp_server.set_server_setting settings="{'MaxRecipients': '500'}"
    '''
    if not settings:
        _LOG.warning('No settings provided')
        return False

    # Some fields are formatted like '{data}'. Salt tries to convert these to dicts
    # automatically on input, so convert them back to the proper format.
    settings = _normalize_server_settings(**settings)

    current_settings = get_server_setting(settings=settings.keys(), server=server)

    if settings == current_settings:
        _LOG.debug('Settings already contain the provided values.')
        return True

    # Note that we must fetch all properties of IIsSmtpServerSetting below, since
    # filtering for specific properties and then attempting to set them will cause
    # an error like: wmi.x_wmi Unexpected COM Error -2147352567
    with salt.utils.winapi.Com():
        try:
            connection = wmi.WMI(namespace=_WMI_NAMESPACE)
            objs = connection.IIsSmtpServerSetting(Name=server)[0]
        except wmi.x_wmi as error:
            _LOG.error('Encountered WMI error: %s', error.com_error)
        except (AttributeError, IndexError) as error:
            _LOG.error('Error getting IIsSmtpServerSetting: %s', error)

        for setting in settings:
            if str(settings[setting]) != str(current_settings[setting]):
                try:
                    setattr(objs, setting, settings[setting])
                except wmi.x_wmi as error:
                    _LOG.error('Encountered WMI error: %s', error.com_error)
                except AttributeError as error:
                    _LOG.error('Error setting %s: %s', setting, error)

    # Get the settings post-change so that we can verify tht all properties
    # were modified successfully. Track the ones that weren't.
    new_settings = get_server_setting(settings=settings.keys(), server=server)
    failed_settings = dict()

    for setting in settings:
        if str(settings[setting]) != str(new_settings[setting]):
            failed_settings[setting] = settings[setting]
    if failed_settings:
        _LOG.error('Failed to change settings: %s', failed_settings)
        return False

    _LOG.debug('Settings configured successfully: %s', settings.keys())
    return True


def get_log_format(server=_DEFAULT_SERVER):
    '''
    Get the active log format for the SMTP virtual server.

    :param str server: The SMTP server name.

    :return: A string of the log format name.
    :rtype: str

    CLI Example:

    .. code-block:: bash

        salt '*' win_smtp_server.get_log_format
    '''
    log_format_types = get_log_format_types()
    format_id = _get_wmi_setting('IIsSmtpServerSetting', 'LogPluginClsid', server)

    # Since IIsSmtpServerSetting stores the log type as an id, we need
    # to get the mapping from IISLogModuleSetting and extract the name.
    for key in log_format_types:
        if str(format_id) == log_format_types[key]:
            return key
    _LOG.warning('Unable to determine log format.')
    return None


def set_log_format(log_format, server=_DEFAULT_SERVER):
    '''
    Set the active log format for the SMTP virtual server.

    :param str log_format: The log format name.
    :param str server: The SMTP server name.

    :return: A boolean representing whether the change succeeded.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' win_smtp_server.set_log_format 'Microsoft IIS Log File Format'
    '''
    setting = 'LogPluginClsid'
    log_format_types = get_log_format_types()
    format_id = log_format_types.get(log_format, None)

    if not format_id:
        message = ("Invalid log format '{0}' specified. Valid formats:"
                   ' {1}').format(log_format, log_format_types.keys())
        raise SaltInvocationError(message)

    _LOG.debug("Id for '%s' found: %s", log_format, format_id)

    current_log_format = get_log_format(server)

    if log_format == current_log_format:
        _LOG.debug('%s already contains the provided format.', setting)
        return True

    _set_wmi_setting('IIsSmtpServerSetting', setting, format_id, server)

    new_log_format = get_log_format(server)
    ret = log_format == new_log_format

    if ret:
        _LOG.debug("Setting %s configured successfully: %s", setting, log_format)
    else:
        _LOG.error("Unable to configure %s with value: %s", setting, log_format)
    return ret


def get_connection_ip_list(as_wmi_format=False, server=_DEFAULT_SERVER):
    '''
    Get the IPGrant list for the SMTP virtual server.

    :param bool as_wmi_format: Returns the connection IPs as a list in the format WMI expects.
    :param str server: The SMTP server name.

    :return: A dictionary of the IP and subnet pairs.
    :rtype: dict

    CLI Example:

    .. code-block:: bash

        salt '*' win_smtp_server.get_connection_ip_list
    '''
    ret = dict()
    setting = 'IPGrant'
    reg_separator = r',\s*'

    if as_wmi_format:
        ret = list()

    addresses = _get_wmi_setting('IIsIPSecuritySetting', setting, server)

    # WMI returns the addresses as a tuple of unicode strings, each representing
    # an address/subnet pair. Remove extra spaces that may be present.
    for unnormalized_address in addresses:
        ip_address, subnet = re.split(reg_separator, unnormalized_address)
        if as_wmi_format:
            ret.append('{0}, {1}'.format(ip_address, subnet))
        else:
            ret[ip_address] = subnet

    if not ret:
        _LOG.debug('%s is empty.', setting)
    return ret


def set_connection_ip_list(addresses=None, grant_by_default=False, server=_DEFAULT_SERVER):
    '''
    Set the IPGrant list for the SMTP virtual server.

    :param str addresses: A dictionary of IP + subnet pairs.
    :param bool grant_by_default: Whether the addresses should be a blacklist or whitelist.
    :param str server: The SMTP server name.

    :return: A boolean representing whether the change succeeded.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' win_smtp_server.set_connection_ip_list addresses="{'127.0.0.1': '255.255.255.255'}"
    '''
    setting = 'IPGrant'
    formatted_addresses = list()

    # It's okay to accept an empty list for set_connection_ip_list,
    # since an empty list may be desirable.
    if not addresses:
        addresses = dict()
        _LOG.debug('Empty %s specified.', setting)

    # Convert addresses to the 'ip_address, subnet' format used by
    # IIsIPSecuritySetting.
    for address in addresses:
        formatted_addresses.append('{0}, {1}'.format(address.strip(),
                                                     addresses[address].strip()))

    current_addresses = get_connection_ip_list(as_wmi_format=True, server=server)

    # Order is not important, so compare to the current addresses as unordered sets.
    if set(formatted_addresses) == set(current_addresses):
        _LOG.debug('%s already contains the provided addresses.', setting)
        return True

    # First we should check GrantByDefault, and change it if necessary.
    current_grant_by_default = _get_wmi_setting('IIsIPSecuritySetting', 'GrantByDefault', server)

    if grant_by_default != current_grant_by_default:
        _LOG.debug('Setting GrantByDefault to: %s', grant_by_default)
        _set_wmi_setting('IIsIPSecuritySetting', 'GrantByDefault', grant_by_default, server)

    _set_wmi_setting('IIsIPSecuritySetting', setting, formatted_addresses, server)

    new_addresses = get_connection_ip_list(as_wmi_format=True, server=server)
    ret = set(formatted_addresses) == set(new_addresses)

    if ret:
        _LOG.debug('%s configured successfully: %s', setting, formatted_addresses)
        return ret
    _LOG.error('Unable to configure %s with value: %s', setting, formatted_addresses)
    return ret


def get_relay_ip_list(server=_DEFAULT_SERVER):
    '''
    Get the RelayIpList list for the SMTP virtual server.

    :param str server: The SMTP server name.

    :return: A list of the relay IPs.
    :rtype: list

    .. note::

        A return value of None corresponds to the restrictive 'Only the list below' GUI parameter
        with an empty access list, and setting an empty list/tuple corresponds to the more
        permissive 'All except the list below' GUI parameter.

    CLI Example:

    .. code-block:: bash

        salt '*' win_smtp_server.get_relay_ip_list
    '''
    ret = list()
    setting = 'RelayIpList'

    lines = _get_wmi_setting('IIsSmtpServerSetting', setting, server)

    if not lines:
        _LOG.debug('%s is empty: %s', setting, lines)
        if lines is None:
            lines = [None]
        return list(lines)

    # WMI returns the addresses as a tuple of individual octets, so we
    # need to group them and reassemble them into IP addresses.
    i = 0
    while i < len(lines):
        octets = [str(x) for x in lines[i: i + 4]]
        address = '.'.join(octets)
        ret.append(address)
        i += 4
    return ret


def set_relay_ip_list(addresses=None, server=_DEFAULT_SERVER):
    '''
    Set the RelayIpList list for the SMTP virtual server.

    Due to the unusual way that Windows stores the relay IPs, it is advisable to retrieve
    the existing list you wish to set from a pre-configured server.

    For example, setting '127.0.0.1' as an allowed relay IP through the GUI would generate
    an actual relay IP list similar to the following:

    .. code-block:: cfg

        ['24.0.0.128', '32.0.0.128', '60.0.0.128', '68.0.0.128', '1.0.0.0', '76.0.0.0',
         '0.0.0.0', '0.0.0.0', '1.0.0.0', '1.0.0.0', '2.0.0.0', '2.0.0.0', '4.0.0.0',
         '0.0.0.0', '76.0.0.128', '0.0.0.0', '0.0.0.0', '0.0.0.0', '0.0.0.0',
         '255.255.255.255', '127.0.0.1']

    .. note::

        Setting the list to None corresponds to the restrictive 'Only the list below' GUI parameter
        with an empty access list configured, and setting an empty list/tuple corresponds to the
        more permissive 'All except the list below' GUI parameter.

    :param str addresses: A list of the relay IPs. The order of the list is important.
    :param str server: The SMTP server name.

    :return: A boolean representing whether the change succeeded.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' win_smtp_server.set_relay_ip_list addresses="['192.168.1.1', '172.16.1.1']"
    '''
    setting = 'RelayIpList'
    formatted_addresses = list()

    current_addresses = get_relay_ip_list(server)

    if list(addresses) == current_addresses:
        _LOG.debug('%s already contains the provided addresses.', setting)
        return True

    if addresses:
        # The WMI input data needs to be in the format used by RelayIpList. Order
        # is also important due to the way RelayIpList orders the address list.

        if addresses[0] is None:
            formatted_addresses = None
        else:
            for address in addresses:
                for octet in address.split('.'):
                    formatted_addresses.append(octet)

    _LOG.debug('Formatted %s addresses: %s', setting, formatted_addresses)

    _set_wmi_setting('IIsSmtpServerSetting', setting, formatted_addresses, server)

    new_addresses = get_relay_ip_list(server)

    ret = list(addresses) == new_addresses

    if ret:
        _LOG.debug('%s configured successfully: %s', setting, addresses)
        return ret
    _LOG.error('Unable to configure %s with value: %s', setting, addresses)
    return ret



# -*- coding: utf-8 -*-
'''
Module for managing the Salt beacons on a minion

.. versionadded:: 2015.8.0

'''

# Import Python libs
from __future__ import absolute_import
import difflib
import logging
import os
import yaml

# Import Salt libs
import salt.utils
import salt.utils.event
from salt.ext.six.moves import map

# Get logging started
log = logging.getLogger(__name__)

__func_alias__ = {
    'list_': 'list',
    'reload_': 'reload'
}


def list_(return_yaml=True):
    '''
    List the beacons currently configured on the minion

    :param return_yaml:     Whether to return YAML formatted output, default True
    :return:                List of currently configured Beacons.

    CLI Example:

    .. code-block:: bash

        salt '*' beacons.list

    '''
    beacons = None

    try:
        eventer = salt.utils.event.get_event('minion', opts=__opts__)
        res = __salt__['event.fire']({'func': 'list'}, 'manage_beacons')
        if res:
            event_ret = eventer.get_event(tag='/salt/minion/minion_beacons_list_complete', wait=30)
            log.debug('event_ret {0}'.format(event_ret))
            if event_ret and event_ret['complete']:
                beacons = event_ret['beacons']
    except KeyError:
        # Effectively a no-op, since we can't really return without an event system
        ret = {}
        ret['result'] = False
        ret['comment'] = 'Event module not available. Beacon add failed.'
        return ret

    if beacons:
        if return_yaml:
            tmp = {'beacons': beacons}
            yaml_out = yaml.safe_dump(tmp, default_flow_style=False)
            return yaml_out
        else:
            return beacons
    else:
        return {'beacons': {}}


def add(name, beacon_data, **kwargs):
    '''
    Add a beacon on the minion

    :param name:            Name of the beacon to configure
    :param beacon_data:     Dictionary or list containing configuration for beacon.
    :return:                Boolean and status message on success or failure of add.

    CLI Example:

    .. code-block:: bash

        salt '*' beacons.add ps "{'salt-master': 'stopped', 'apache2': 'stopped'}"

    '''
    ret = {'comment': 'Failed to add beacon {0}.'.format(name),
           'result': False}

    if name in list_(return_yaml=False):
        ret['comment'] = 'Beacon {0} is already configured.'.format(name)
        return ret

    if 'test' in kwargs and kwargs['test']:
        ret['result'] = True
        ret['comment'] = 'Beacon: {0} would be added.'.format(name)
    else:
        # Attempt to load the beacon module so we have access to the validate function
        try:
            beacon_module = __import__('salt.beacons.' + name, fromlist=['validate'])
            log.debug('Successfully imported beacon.')
        except ImportError:
            ret['comment'] = 'Beacon {0} does not exist'.format(name)
            return ret

        # Attempt to validate
        if hasattr(beacon_module, 'validate'):
            _beacon_data = beacon_data
            if 'enabled' in _beacon_data:
                del _beacon_data['enabled']
            valid, vcomment = beacon_module.validate(_beacon_data)
        else:
            log.info('Beacon {0} does not have a validate'
                     ' function,  skipping validation.'.format(name))
            valid = True

        if not valid:
            ret['result'] = False
            ret['comment'] = ('Beacon {0} configuration invalid, '
                              'not adding.\n{1}'.format(name, vcomment))
            return ret

        try:
            eventer = salt.utils.event.get_event('minion', opts=__opts__)
            res = __salt__['event.fire']({'name': name, 'beacon_data': beacon_data, 'func': 'add'}, 'manage_beacons')
            if res:
                event_ret = eventer.get_event(tag='/salt/minion/minion_beacon_add_complete', wait=30)
                if event_ret and event_ret['complete']:
                    beacons = event_ret['beacons']
                    if name in beacons and beacons[name] == beacon_data:
                        ret['result'] = True
                        ret['comment'] = 'Added beacon: {0}.'.format(name)
                        return ret
        except KeyError:
            # Effectively a no-op, since we can't really return without an event system
            ret['comment'] = 'Event module not available. Beacon add failed.'
    return ret


def modify(name, beacon_data, **kwargs):
    '''
    Modify an existing beacon

    :param name:            Name of the beacon to configure
    :param beacon_data:     Dictionary or list containing updated configuration for beacon.
    :return:                Boolean and status message on success or failure of modify.

    CLI Example:

    .. code-block:: bash

        salt '*' beacons.modify ps "{'salt-master': 'stopped', 'apache2': 'stopped'}"
    '''

    ret = {'comment': '',
           'result': True}

    current_beacons = list_(return_yaml=False)
    if name not in current_beacons:
        ret['comment'] = 'Beacon {0} is not configured.'.format(name)
        return ret

    if 'test' in kwargs and kwargs['test']:
        ret['result'] = True
        ret['comment'] = 'Beacon: {0} would be added.'.format(name)
    else:
        # Attempt to load the beacon module so we have access to the validate function
        try:
            beacon_module = __import__('salt.beacons.' + name, fromlist=['validate'])
            log.debug('Successfully imported beacon.')
        except ImportError:
            ret['comment'] = 'Beacon {0} does not exist'.format(name)
            return ret

        # Attempt to validate
        if hasattr(beacon_module, 'validate'):
            _beacon_data = beacon_data
            if 'enabled' in _beacon_data:
                del _beacon_data['enabled']
            valid, vcomment = beacon_module.validate(_beacon_data)
        else:
            log.info('Beacon {0} does not have a validate'
                     ' function,  skipping validation.'.format(name))
            valid = True

        if not valid:
            ret['result'] = False
            ret['comment'] = ('Beacon {0} configuration invalid, '
                              'not adding.\n{1}'.format(name, vcomment))
            return ret

        _current = current_beacons[name]
        _new = beacon_data

        if _new == _current:
            ret['comment'] = 'Job {0} in correct state'.format(name)
            return ret

        _current_lines = ['{0}:{1}\n'.format(key, value)
                          for (key, value) in sorted(_current.items())]
        _new_lines = ['{0}:{1}\n'.format(key, value)
                      for (key, value) in sorted(_new.items())]
        _diff = difflib.unified_diff(_current_lines, _new_lines)

        ret['changes'] = {}
        ret['changes']['diff'] = ''.join(_diff)

        try:
            eventer = salt.utils.event.get_event('minion', opts=__opts__)
            res = __salt__['event.fire']({'name': name, 'beacon_data': beacon_data, 'func': 'modify'}, 'manage_beacons')
            if res:
                event_ret = eventer.get_event(tag='/salt/minion/minion_beacon_modify_complete', wait=30)
                if event_ret and event_ret['complete']:
                    beacons = event_ret['beacons']
                    if name in beacons and beacons[name] == beacon_data:
                        ret['result'] = True
                        ret['comment'] = 'Modified beacon: {0}.'.format(name)
                        return ret
        except KeyError:
            # Effectively a no-op, since we can't really return without an event system
            ret['comment'] = 'Event module not available. Beacon add failed.'
    return ret


def delete(name, **kwargs):
    '''
    Delete a beacon item

    :param name:            Name of the beacon to delete
    :return:                Boolean and status message on success or failure of delete.

    CLI Example:

    .. code-block:: bash

        salt '*' beacons.delete ps

        salt '*' beacons.delete load

    '''

    ret = {'comment': 'Failed to delete beacon {0}.'.format(name),
           'result': False}

    if 'test' in kwargs and kwargs['test']:
        ret['result'] = True
        ret['comment'] = 'Beacon: {0} would be deleted.'.format(name)
    else:
        try:
            eventer = salt.utils.event.get_event('minion', opts=__opts__)
            res = __salt__['event.fire']({'name': name, 'func': 'delete'}, 'manage_beacons')
            if res:
                event_ret = eventer.get_event(tag='/salt/minion/minion_beacon_delete_complete', wait=30)
                if event_ret and event_ret['complete']:
                    beacons = event_ret['beacons']
                    if name not in beacons:
                        ret['result'] = True
                        ret['comment'] = 'Deleted beacon: {0}.'.format(name)
                        return ret
        except KeyError:
            # Effectively a no-op, since we can't really return without an event system
            ret['comment'] = 'Event module not available. Beacon add failed.'
    return ret


def save():
    '''
    Save all beacons on the minion

    :return:                Boolean and status message on success or failure of save.

    CLI Example:

    .. code-block:: bash

        salt '*' beacons.save
    '''

    ret = {'comment': [],
           'result': True}

    beacons = list_(return_yaml=False)

    # move this file into an configurable opt
    sfn = '{0}/{1}/beacons.conf'.format(__opts__['config_dir'],
                                        os.path.dirname(__opts__['default_include']))
    if beacons:
        tmp = {'beacons': beacons}
        yaml_out = yaml.safe_dump(tmp, default_flow_style=False)
    else:
        yaml_out = ''

    try:
        with salt.utils.fopen(sfn, 'w+') as fp_:
            fp_.write(yaml_out)
        ret['comment'] = 'Beacons saved to {0}.'.format(sfn)
    except (IOError, OSError):
        ret['comment'] = 'Unable to write to beacons file at {0}. Check permissions.'.format(sfn)
        ret['result'] = False
    return ret


def enable(**kwargs):
    '''
    Enable all beacons on the minion

    :return:                Boolean and status message on success or failure of enable.

    CLI Example:

    .. code-block:: bash

        salt '*' beacons.enable
    '''

    ret = {'comment': [],
           'result': True}

    if 'test' in kwargs and kwargs['test']:
        ret['comment'] = 'Beacons would be enabled.'
    else:
        try:
            eventer = salt.utils.event.get_event('minion', opts=__opts__)
            res = __salt__['event.fire']({'func': 'enable'}, 'manage_beacons')
            if res:
                event_ret = eventer.get_event(tag='/salt/minion/minion_beacons_enabled_complete', wait=30)
                if event_ret and event_ret['complete']:
                    beacons = event_ret['beacons']
                    if 'enabled' in beacons and beacons['enabled']:
                        ret['result'] = True
                        ret['comment'] = 'Enabled beacons on minion.'
                    else:
                        ret['result'] = False
                        ret['comment'] = 'Failed to enable beacons on minion.'
                    return ret
        except KeyError:
            # Effectively a no-op, since we can't really return without an event system
            ret['comment'] = 'Event module not available. Beacons enable job failed.'
    return ret


def disable(**kwargs):
    '''
    Disable all beaconsd jobs on the minion

    :return:                Boolean and status message on success or failure of disable.

    CLI Example:

    .. code-block:: bash

        salt '*' beacons.disable
    '''

    ret = {'comment': [],
           'result': True}

    if 'test' in kwargs and kwargs['test']:
        ret['comment'] = 'Beacons would be disabled.'
    else:
        try:
            eventer = salt.utils.event.get_event('minion', opts=__opts__)
            res = __salt__['event.fire']({'func': 'disable'}, 'manage_beacons')
            if res:
                event_ret = eventer.get_event(tag='/salt/minion/minion_beacons_disabled_complete', wait=30)
                log.debug('event_ret {0}'.format(event_ret))
                if event_ret and event_ret['complete']:
                    beacons = event_ret['beacons']
                    if 'enabled' in beacons and not beacons['enabled']:
                        ret['result'] = True
                        ret['comment'] = 'Disabled beacons on minion.'
                    else:
                        ret['result'] = False
                        ret['comment'] = 'Failed to disable beacons on minion.'
                    return ret
        except KeyError:
            # Effectively a no-op, since we can't really return without an event system
            ret['comment'] = 'Event module not available. Beacons enable job failed.'
    return ret


def _get_beacon_config_dict(beacon_config):
    beacon_config_dict = {}
    if isinstance(beacon_config, list):
        list(map(beacon_config_dict.update, beacon_config))
    else:
        beacon_config_dict = beacon_config

    return beacon_config_dict


def enable_beacon(name, **kwargs):
    '''
    Enable beacon on the minion

    :name:                  Name of the beacon to enable.
    :return:                Boolean and status message on success or failure of enable.

    CLI Example:

    .. code-block:: bash

        salt '*' beacons.enable_beacon ps
    '''

    ret = {'comment': [],
           'result': True}

    if not name:
        ret['comment'] = 'Beacon name is required.'
        ret['result'] = False
        return ret

    if 'test' in kwargs and kwargs['test']:
        ret['comment'] = 'Beacon {0} would be enabled.'.format(name)
    else:
        _beacons = list_(return_yaml=False)
        if name not in _beacons:
            ret['comment'] = 'Beacon {0} is not currently configured.'.format(name)
            ret['result'] = False
            return ret

        try:
            eventer = salt.utils.event.get_event('minion', opts=__opts__)
            res = __salt__['event.fire']({'func': 'enable_beacon', 'name': name}, 'manage_beacons')
            if res:
                event_ret = eventer.get_event(tag='/salt/minion/minion_beacon_enabled_complete', wait=30)
                if event_ret and event_ret['complete']:
                    beacons = event_ret['beacons']
                    beacon_config_dict = _get_beacon_config_dict(beacons[name])

                    if 'enabled' in beacon_config_dict and beacon_config_dict['enabled']:
                        ret['result'] = True
                        ret['comment'] = 'Enabled beacon {0} on minion.'.format(name)
                    else:
                        ret['result'] = False
                        ret['comment'] = 'Failed to enable beacon {0} on minion.'.format(name)
                    return ret
        except KeyError:
            # Effectively a no-op, since we can't really return without an event system
            ret['comment'] = 'Event module not available. Beacon enable job failed.'
    return ret


def disable_beacon(name, **kwargs):
    '''
    Disable beacon on the minion

    :name:                  Name of the beacon to disable.
    :return:                Boolean and status message on success or failure of disable.

    CLI Example:

    .. code-block:: bash

        salt '*' beacons.disable_beacon ps
    '''

    ret = {'comment': [],
           'result': True}

    if not name:
        ret['comment'] = 'Beacon name is required.'
        ret['result'] = False
        return ret

    if 'test' in kwargs and kwargs['test']:
        ret['comment'] = 'Beacons would be enabled.'
    else:
        _beacons = list_(return_yaml=False)
        if name not in _beacons:
            ret['comment'] = 'Beacon {0} is not currently configured.'.format(name)
            ret['result'] = False
            return ret

        try:
            eventer = salt.utils.event.get_event('minion', opts=__opts__)
            res = __salt__['event.fire']({'func': 'disable_beacon', 'name': name}, 'manage_beacons')
            if res:
                event_ret = eventer.get_event(tag='/salt/minion/minion_beacon_disabled_complete', wait=30)
                if event_ret and event_ret['complete']:
                    beacons = event_ret['beacons']
                    beacon_config_dict = _get_beacon_config_dict(beacons[name])

                    if 'enabled' in beacon_config_dict and not beacon_config_dict['enabled']:
                        ret['result'] = True
                        ret['comment'] = 'Disabled beacon {0} on minion.'.format(name)
                    else:
                        ret['result'] = False
                        ret['comment'] = 'Failed to disable beacon on minion.'
                    return ret
        except KeyError:
            # Effectively a no-op, since we can't really return without an event system
            ret['comment'] = 'Event module not available. Beacon disable job failed.'
    return ret



# -*- coding: utf-8 -*-
'''
Splay function calls across targeted minions

@author: Dmitry Kuzmenko <dmitry.kuzmenko@dsr-company.com>
'''
# Import Python Libs
from __future__ import absolute_import
import time
import logging

from salt.executors import ModuleExecutorBase

log = logging.getLogger(__name__)

_DEFAULT_SPLAYTIME = 300
_DEFAULT_SIZE = 8192


def get(*args, **kwargs):
    return SplayExecutor(*args, **kwargs)


class SplayExecutor(ModuleExecutorBase):
    '''
    classdocs
    '''

    def __init__(self, opts, data, executor):
        '''
        Constructor
        '''
        super(SplayExecutor, self).__init__()
        self.splaytime = data.get('splaytime') or opts.get('splaytime', _DEFAULT_SPLAYTIME)
        if self.splaytime <= 0:
            raise ValueError('splaytime must be a positive integer')
        self.executor = executor
        self.fun_name = data.get('fun')

    def _get_hash(self, hashable, size):
        '''
        Jenkins One-At-A-Time Hash Function
        More Info: http://en.wikipedia.org/wiki/Jenkins_hash_function#one-at-a-time
        '''
        # Using bitmask to emulate rollover behavior of C unsigned 32 bit int
        bitmask = 0xffffffff
        h = 0

        for i in bytearray(hashable):
            h = (h + i) & bitmask
            h = (h + (h << 10)) & bitmask
            h = (h ^ (h >> 6)) & bitmask

        h = (h + (h << 3)) & bitmask
        h = (h ^ (h >> 11)) & bitmask
        h = (h + (h << 15)) & bitmask

        return (h & (size - 1)) & bitmask

    def _calc_splay(self, hashable, splaytime=_DEFAULT_SPLAYTIME, size=_DEFAULT_SIZE):
        hash_val = self._get_hash(hashable, size)
        return int(splaytime * hash_val / float(size))

    def execute(self):
        '''
        Splay a salt function call execution time across minions over
        a number of seconds (default: 600)

        .. note::
            You *probably* want to use --async here and look up the job results later.
            If you're dead set on getting the output from the CLI command, then make
            sure to set the timeout (with the -t flag) to something greater than the
            splaytime (max splaytime + time to execute job).
            Otherwise, it's very likely that the cli will time out before the job returns.

        CLI Example:

        .. code-block:: bash

            # With default splaytime
            salt --async '*' splay.splay pkg.install cowsay version=3.03-8.el6

        .. code-block:: bash

            # With specified splaytime (5 minutes) and timeout with 10 second buffer
            salt -t 310 '*' splay.splay 300 pkg.version cowsay
        '''
        my_delay = self._calc_splay(__grains__['id'], splaytime=self.splaytime)
        log.debug("Splay is sleeping {0} secs on {1}".format(my_delay, self.fun_name))
        time.sleep(my_delay)
        return self.executor.execute()



# -*- coding: utf-8 -*-
'''
    :codeauthor: :email:`Jayesh Kariya <jayeshk@saltstack.com>`
'''
# Import Python libs
from __future__ import absolute_import

# Import Salt Testing Libs
from salttesting import skipIf, TestCase
from salttesting.mock import (
    NO_MOCK,
    NO_MOCK_REASON,
    MagicMock,
    patch
)

from salttesting.helpers import ensure_in_syspath

ensure_in_syspath('../../')

# Import Salt Libs
from salt.states import postgres_schema

postgres_schema.__opts__ = {}
postgres_schema.__salt__ = {}


@skipIf(NO_MOCK, NO_MOCK_REASON)
class PostgresSchemaTestCase(TestCase):
    '''
    Test cases for salt.states.postgres_schema
    '''
    # 'present' function tests: 1

    def test_present(self):
        '''
        Test to ensure that the named schema is present in the database.
        '''
        name = 'myname'
        dbname = 'mydb'

        ret = {'name': name,
               'dbname': dbname,
               'changes': {},
               'result': True,
               'comment': ''}

        mock = MagicMock(return_value=name)
        with patch.dict(postgres_schema.__salt__,
                        {'postgres.schema_get': mock}):
            comt = ('Schema {0} already exists in database {1}'.format(name,
                                                                       dbname))
            ret.update({'comment': comt})
            self.assertDictEqual(postgres_schema.present(dbname, name), ret)

    # 'absent' function tests: 1

    def test_absent(self):
        '''
        Test to ensure that the named schema is absent.
        '''
        name = 'myname'
        dbname = 'mydb'

        ret = {'name': name,
               'dbname': dbname,
               'changes': {},
               'result': True,
               'comment': ''}

        mock_t = MagicMock(side_effect=[True, False])
        mock = MagicMock(side_effect=[True, True, False])
        with patch.dict(postgres_schema.__salt__,
                        {'postgres.schema_exists': mock,
                         'postgres.schema_remove': mock_t}):
            comt = ('Schema {0} has been removed from database {1}'.
                    format(name, dbname))
            ret.update({'comment': comt, 'result': True,
                        'changes': {name: 'Absent'}})
            self.assertDictEqual(postgres_schema.absent(dbname, name), ret)

            comt = ('Schema {0} failed to be removed'.format(name))
            ret.update({'comment': comt, 'result': False, 'changes': {}})
            self.assertDictEqual(postgres_schema.absent(dbname, name), ret)

            comt = ('Schema {0} is not present in database {1},'
                    ' so it cannot be removed'.format(name, dbname))
            ret.update({'comment': comt, 'result': True})
            self.assertDictEqual(postgres_schema.absent(dbname, name), ret)


if __name__ == '__main__':
    from integration import run_tests
    run_tests(PostgresSchemaTestCase, needs_daemon=False)



# -*- coding: utf-8 -*-
'''
NAPALM NTP
==========

Manages NTP on network devices.

:codeauthor: Mircea Ulinic <mircea@cloudflare.com> & Jerome Fleury <jf@cloudflare.com>
:maturity:   new
:depends:    napalm
:platform:   unix

Dependencies
------------
- :mod:`NAPALM proxy minion <salt.proxy.napalm>`
- :mod:`NET basic features <salt.modules.napalm_network>`

.. seealso::
    :mod:`NTP peers management state <salt.states.netntp>`

.. versionadded:: 2016.11.0
'''

from __future__ import absolute_import

import logging
log = logging.getLogger(__file__)


try:
    # will try to import NAPALM
    # https://github.com/napalm-automation/napalm
    # pylint: disable=W0611
    from napalm_base import get_network_driver
    # pylint: enable=W0611
    HAS_NAPALM = True
except ImportError:
    HAS_NAPALM = False

# ----------------------------------------------------------------------------------------------------------------------
# module properties
# ----------------------------------------------------------------------------------------------------------------------

__virtualname__ = 'ntp'
__proxyenabled__ = ['napalm']
# uses NAPALM-based proxy to interact with network devices

# ----------------------------------------------------------------------------------------------------------------------
# property functions
# ----------------------------------------------------------------------------------------------------------------------


def __virtual__():

    '''
    NAPALM library must be installed for this module to work.
    Also, the key proxymodule must be set in the __opts___ dictionary.
    '''

    if HAS_NAPALM and 'proxy' in __opts__:
        return __virtualname__
    else:
        return (False, 'The module NTP cannot be loaded: \
                napalm or proxy could not be loaded.')

# ----------------------------------------------------------------------------------------------------------------------
# helper functions -- will not be exported
# ----------------------------------------------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------------------------------------------
# callable functions
# ----------------------------------------------------------------------------------------------------------------------


def peers():

    '''
    Returns a list the NTP peers configured on the network device.

    :return: configured NTP peers as list.

    CLI Example:

    .. code-block:: bash

        salt '*' ntp.peers

    Example output:

    .. code-block:: python

        [
            '192.168.0.1',
            '172.17.17.1',
            '172.17.17.2',
            '2400:cb00:6:1024::c71b:840a'
        ]

    '''

    ntp_peers = __proxy__['napalm.call'](
        'get_ntp_peers',
        **{
        }
    )

    if not ntp_peers.get('result'):
        return ntp_peers

    ntp_peers_list = list(ntp_peers.get('out', {}).keys())

    ntp_peers['out'] = ntp_peers_list

    return ntp_peers


def servers():

    '''
    Returns a list of the configured NTP servers on the device.

    CLI Example:

    .. code-block:: bash

        salt '*' ntp.servers
    '''

    ntp_servers = __proxy__['napalm.call'](
        'get_ntp_servers',
        **{
        }
    )

    if not ntp_servers.get('result'):
        return ntp_servers

    ntp_servers_list = list(ntp_servers.get('out', {}).keys())

    ntp_servers['out'] = ntp_servers_list

    return ntp_servers


def stats(peer=None):

    '''
    Returns a dictionary containing synchronization details of the NTP peers.

    :param peer: Returns only the details of a specific NTP peer.
    :return: a list of dictionaries, with the following keys:

        * remote
        * referenceid
        * synchronized
        * stratum
        * type
        * when
        * hostpoll
        * reachability
        * delay
        * offset
        * jitter

    CLI Example:

    .. code-block:: bash

        salt '*' ntp.stats

    Example output:

    .. code-block:: python

        [
            {
                'remote'        : u'188.114.101.4',
                'referenceid'   : u'188.114.100.1',
                'synchronized'  : True,
                'stratum'       : 4,
                'type'          : u'-',
                'when'          : u'107',
                'hostpoll'      : 256,
                'reachability'  : 377,
                'delay'         : 164.228,
                'offset'        : -13.866,
                'jitter'        : 2.695
            }
        ]
    '''

    proxy_output = __proxy__['napalm.call'](
        'get_ntp_stats',
        **{
        }
    )

    if not proxy_output.get('result'):
        return proxy_output

    ntp_peers = proxy_output.get('out')

    if peer:
        ntp_peers = [ntp_peer for ntp_peer in ntp_peers if ntp_peer.get('remote', '') == peer]

    proxy_output.update({
        'out': ntp_peers
    })

    return proxy_output


def set_peers(*peers, **options):

    '''
    Configures a list of NTP peers on the device.

    :param peers: list of IP Addresses/Domain Names
    :param test (bool): discard loaded config. By default `test` is False (will not dicard the changes)
    :commit commit (bool): commit loaded config. By default `commit` is True (will commit the changes). Useful when
    the user does not want to commit after each change, but after a couple.

    By default this function will commit the config changes (if any). To load without commiting, use the `commit`
    option. For dry run use the `test` argument.

    CLI Example:

    .. code-block:: bash

        salt '*' ntp.set_peers 192.168.0.1 172.17.17.1 time.apple.com
        salt '*' ntp.set_peers 172.17.17.1 test=True  # only displays the diff
        salt '*' ntp.set_peers 192.168.0.1 commit=False  # preserves the changes, but does not commit
    '''

    test = options.pop('test', False)
    commit = options.pop('commit', True)

    return __salt__['net.load_template']('set_ntp_peers',
                                         peers=peers,
                                         test=test,
                                         commit=commit)


def set_servers(*servers, **options):

    '''
    Configures a list of NTP servers on the device.

    :param servers: list of IP Addresses/Domain Names
    :param test (bool): discard loaded config. By default `test` is False (will not dicard the changes)
    :commit commit (bool): commit loaded config. By default `commit` is True (will commit the changes). Useful when
    the user does not want to commit after each change, but after a couple.

    By default this function will commit the config changes (if any). To load without commiting, use the `commit`
    option. For dry run use the `test` argument.

    CLI Example:

    .. code-block:: bash

        salt '*' ntp.set_servers 192.168.0.1 172.17.17.1 time.apple.com
        salt '*' ntp.set_servers 172.17.17.1 test=True  # only displays the diff
        salt '*' ntp.set_servers 192.168.0.1 commit=False  # preserves the changes, but does not commit
    '''

    test = options.pop('test', False)
    commit = options.pop('commit', True)

    return __salt__['net.load_template']('set_ntp_servers',
                                         servers=servers,
                                         test=test,
                                         commit=commit)


def delete_peers(*peers, **options):

    '''
    Removes NTP peers configured on the device.

    :param peers: list of IP Addresses/Domain Names to be removed as NTP peers
    :param test (bool): discard loaded config. By default `test` is False (will not dicard the changes)
    :commit commit (bool): commit loaded config. By default `commit` is True (will commit the changes). Useful when
    the user does not want to commit after each change, but after a couple.

    By default this function will commit the config changes (if any). To load without commiting, use the `commit`
    option. For dry run use the `test` argument.

    CLI Example:

    .. code-block:: bash

        salt '*' ntp.delete_peers 8.8.8.8 time.apple.com
        salt '*' ntp.delete_peers 172.17.17.1 test=True  # only displays the diff
        salt '*' ntp.delete_peers 192.168.0.1 commit=False  # preserves the changes, but does not commit
    '''

    test = options.pop('test', False)
    commit = options.pop('commit', True)

    return __salt__['net.load_template']('delete_ntp_peers',
                                         peers=peers,
                                         test=test,
                                         commit=commit)


def delete_servers(*servers, **options):

    '''
    Removes NTP servers configured on the device.

    :param servers: list of IP Addresses/Domain Names to be removed as NTP servers
    :param test (bool): discard loaded config. By default `test` is False (will not dicard the changes)
    :commit commit (bool): commit loaded config. By default `commit` is True (will commit the changes). Useful when
    the user does not want to commit after each change, but after a couple.

    By default this function will commit the config changes (if any). To load without commiting, use the `commit`
    option. For dry run use the `test` argument.

    CLI Example:

    .. code-block:: bash

        salt '*' ntp.delete_servers 8.8.8.8 time.apple.com
        salt '*' ntp.delete_servers 172.17.17.1 test=True  # only displays the diff
        salt '*' ntp.delete_servers 192.168.0.1 commit=False  # preserves the changes, but does not commit
    '''

    test = options.pop('test', False)
    commit = options.pop('commit', True)

    return __salt__['net.load_template']('delete_ntp_servers',
                                         servers=servers,
                                         test=test,
                                         commit=commit)



# -*- coding: utf-8 -*-
'''
Module for managing PowerShell through PowerShellGet (PSGet)

:depends:
    - PowerShell 5.0
    - PSGet

Support for PowerShell
'''
from __future__ import absolute_import, unicode_literals

# Import python libs
import copy
import logging
import json

# Import salt libs
import salt.utils
from salt.exceptions import CommandExecutionError

# Set up logging
log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = 'psget'


def __virtual__():
    '''
    Set the system module of the kernel is Windows
    '''
    # Verify Windows
    if not salt.utils.is_windows():
        log.debug('Module PSGet: Only available on Windows systems')
        return False, 'Module PSGet: Only available on Windows systems'

    # Verify PowerShell
    powershell_info = __salt__['cmd.shell_info']('powershell')
    if not powershell_info['installed']:
        log.debug('Module PSGet: Requires PowerShell')
        return False, 'Module PSGet: Requires PowerShell'

    # Verify PowerShell 5.0 or greater
    if salt.utils.compare_versions(powershell_info['version'], '<', '5.0'):
        log.debug('Module PSGet: Requires PowerShell 5 or newer')
        return False, 'Module PSGet: Requires PowerShell 5 or newer.'

    return __virtualname__


def _pshell(cmd, cwd=None, json_depth=2):
    '''
    Execute the desired powershell command and ensure that it returns data
    in json format and load that into python
    '''
    if 'convertto-json' not in cmd.lower():
        cmd = '{0} | ConvertTo-Json -Depth {1}'.format(cmd, json_depth)
    log.debug('DSC: {0}'.format(cmd))
    results = __salt__['cmd.run_all'](cmd, shell='powershell', cwd=cwd, python_shell=True)

    if 'pid' in results:
        del results['pid']

    if 'retcode' not in results or results['retcode'] != 0:
        # run_all logs an error to log.error, fail hard back to the user
        raise CommandExecutionError('Issue executing powershell {0}'.format(cmd), info=results)

    try:
        ret = json.loads(results['stdout'], strict=False)
    except ValueError:
        raise CommandExecutionError('No JSON results from powershell', info=results)

    return ret


def psversion():
    '''
    Returns the Powershell version

    This has been deprecated and has been replaced by ``cmd.shell_info`` Note
    the minimum version return is 5 as ``dsc`` is not available for version
    less than 5.  This function will be removed in 'Oxygen' release.

    CLI Example:

    .. code-block:: bash

        salt 'win01' dsc.psversion
    '''
    salt.utils.warn_until('Oxygen',
        'The \'psversion\' has been deprecated and has been '
        'replaced by \'cmd.shell_info\'.'
    )
    powershell_info = __salt__['cmd.shell_info']('powershell')
    if powershell_info['installed']:
        try:
            return int(powershell_info['version'].split('.')[0])
        except ValueError:
            pass
    return 0


def bootstrap():
    '''
    Make sure that nuget-anycpu.exe is installed.
    This will download the official nuget-anycpu.exe from the internet.

    CLI Example:

    .. code-block:: bash

        salt 'win01' psget.bootstrap
    '''
    cmd = 'Get-PackageProvider -Name NuGet -ForceBootstrap'
    ret = _pshell(cmd)
    return ret


def avail_modules(desc=False):
    '''
    List available modules in registered Powershell module repositories.

    :param desc: If ``True``, the verbose description will be returned.
    :type  desc: ``bool``

    CLI Example:

    .. code-block:: bash

        salt 'win01' psget.avail_modules
        salt 'win01' psget.avail_modules desc=True
    '''
    cmd = 'Find-Module'
    modules = _pshell(cmd)
    names = []
    if desc:
        names = {}
    for module in modules:
        if desc:
            names[module['Name']] = module['Description']
            continue
        names.append(module['Name'])
    return names


def list_modules(desc=False):
    '''
    List currently installed PSGet Modules on the system.

    :param desc: If ``True``, the verbose description will be returned.
    :type  desc: ``bool``

    CLI Example:

    .. code-block:: bash

        salt 'win01' psget.list_modules
        salt 'win01' psget.list_modules desc=True
    '''
    cmd = 'Get-InstalledModule'
    modules = _pshell(cmd)
    if isinstance(modules, dict):
        ret = []
        if desc:
            modules_ret = {}
            modules_ret[modules['Name']] = copy.deepcopy(modules)
            modules = modules_ret
            return modules
        ret.append(modules['Name'])
        return ret
    names = []
    if desc:
        names = {}
    for module in modules:
        if desc:
            names[module['Name']] = module
            continue
        names.append(module['Name'])
    return names


def install(name, minimum_version=None, required_version=None, scope=None,
            repository=None):
    '''
    Install a Powershell module from powershell gallery on the system.

    :param name: Name of a Powershell module
    :type  name: ``str``

    :param minimum_version: The maximum version to install, e.g. 1.23.2
    :type  minimum_version: ``str``

    :param required_version: Install a specific version
    :type  required_version: ``str``

    :param scope: The scope to install the module to, e.g. CurrentUser, Computer
    :type  scope: ``str``

    :param repository: The friendly name of a private repository, e.g. MyREpo
    :type  repository: ``str``

    CLI Example:

    .. code-block:: bash

        salt 'win01' psget.install PowerPlan
    '''
    # Putting quotes around the parameter protects against command injection
    flags = [('Name', name)]

    if minimum_version is not None:
        flags.append(('MinimumVersion', minimum_version))
    if required_version is not None:
        flags.append(('RequiredVersion', required_version))
    if scope is not None:
        flags.append(('Scope', scope))
    if repository is not None:
        flags.append(('Repository', repository))
    params = ''
    for flag, value in flags:
        params += '-{0} {1} '.format(flag, value)
    cmd = 'Install-Module {0} -Force'.format(params)
    _pshell(cmd)
    return name in list_modules()


def update(name, maximum_version=None, required_version=None):
    '''
    Update a PowerShell module to a specific version, or the newest

    :param name: Name of a Powershell module
    :type  name: ``str``

    :param maximum_version: The maximum version to install, e.g. 1.23.2
    :type  maximum_version: ``str``

    :param required_version: Install a specific version
    :type  required_version: ``str``

    CLI Example:

    .. code-block:: bash

        salt 'win01' psget.update PowerPlan
    '''
    # Putting quotes around the parameter protects against command injection
    flags = [('Name', name)]

    if maximum_version is not None:
        flags.append(('MaximumVersion', maximum_version))
    if required_version is not None:
        flags.append(('RequiredVersion', required_version))

    params = ''
    for flag, value in flags:
        params += '-{0} {1} '.format(flag, value)
    cmd = 'Update-Module {0} -Force'.format(params)
    _pshell(cmd)
    return name in list_modules()


def remove(name):
    '''
    Remove a Powershell DSC module from the system.

    :param  name: Name of a Powershell DSC module
    :type   name: ``str``

    CLI Example:

    .. code-block:: bash

        salt 'win01' psget.remove PowerPlan
    '''
    # Putting quotes around the parameter protects against command injection
    cmd = 'Uninstall-Module "{0}"'.format(name)
    no_ret = _pshell(cmd)
    return name not in list_modules()


def register_repository(name, location, installation_policy=None):
    '''
    Register a PSGet repository on the local machine

    :param name: The name for the repository
    :type  name: ``str``

    :param location: The URI for the repository
    :type  location: ``str``

    :param installation_policy: The installation policy
        for packages, e.g. Trusted, Untrusted
    :type  installation_policy: ``str``

    CLI Example:

    .. code-block:: bash

        salt 'win01' psget.register_repository MyRepo https://myrepo.mycompany.com/packages
    '''
    # Putting quotes around the parameter protects against command injection
    flags = [('Name', name)]

    flags.append(('SourceLocation', location))
    if installation_policy is not None:
        flags.append(('InstallationPolicy', installation_policy))

    params = ''
    for flag, value in flags:
        params += '-{0} {1} '.format(flag, value)
    cmd = 'Register-PSRepository {0}'.format(params)
    no_ret = _pshell(cmd)
    return name not in list_modules()


def get_repository(name):
    '''
    Get the details of a local PSGet repository

    :param  name: Name of the repository
    :type   name: ``str``

    CLI Example:

    .. code-block:: bash

        salt 'win01' psget.get_repository MyRepo
    '''
    # Putting quotes around the parameter protects against command injection
    cmd = 'Get-PSRepository "{0}"'.format(name)
    no_ret = _pshell(cmd)
    return name not in list_modules()



# -*- coding: utf-8 -*-
"""
Management of SQLite3 databases
===============================

:depends:   - SQLite3 Python Module
:configuration: See :py:mod:`salt.modules.sqlite3` for setup instructions
.. versionadded:: 2016.3.0

The sqlite3 module is used to create and manage sqlite3 databases
and execute queries

Here is an example of creating a table using sql statements:

  .. code-block:: yaml

    users:
      sqlite3.table_present:
        - db: /var/www/data/app.sqlite
        - schema: CREATE TABLE `users` (`username` TEXT COLLATE NOCASE UNIQUE NOT NULL, `password` BLOB NOT NULL, `salt` BLOB NOT NULL, `last_login` INT)


Here is an example of creating a table using yaml/jinja instead of sql:

  .. code-block:: yaml

    users:
      sqlite3.table_present:
        - db: /var/www/app.sqlite
        - schema:
          - email TEXT COLLATE NOCASE UNIQUE NOT NULL
          - firstname TEXT NOT NULL
          - lastname TEXT NOT NULL
          - company TEXT NOT NULL
          - password BLOB NOT NULL
          - salt BLOB NOT NULL


Here is an example of making sure a table is absent:

  .. code-block:: yaml

    badservers:
      sqlite3.table_absent:
        - db: /var/www/data/users.sqlite


Sometimes you would to have specific data in tables to be used by other services
Here is an example of making sure rows with specific data exist:

  .. code-block:: yaml

    user_john_doe_xyz:
      sqlite3.row_present:
        - db: /var/www/app.sqlite
        - table: users
        - where_sql: email='john.doe@companyxyz.com'
        - data:
            email: john.doe@companyxyz.com
            lastname: doe
            firstname: john
            company: companyxyz.com
            password: abcdef012934125
            salt: abcdef012934125
        - require:
          - sqlite3: users


Here is an example of removing a row from a table:

  .. code-block:: yaml

    user_john_doe_abc:
      sqlite3.row_absent:
        - db: /var/www/app.sqlite
        - table: users
        - where_sql: email="john.doe@companyabc.com"
        - require:
          - sqlite3: users
"""

# Import Python libs
from __future__ import absolute_import

# Import Salt libs
import salt.ext.six as six

try:
    import sqlite3

    HAS_SQLITE3 = True
except ImportError:
    HAS_SQLITE3 = False


def __virtual__():
    """
    Only load if the sqlite3 module is available
    """
    return HAS_SQLITE3


def row_absent(name, db, table, where_sql, where_args=None):
    """
    Makes sure the specified row is absent in db.  If multiple rows
    match where_sql, then the state will fail.

    name
        Only used as the unique ID

    db
        The database file name

    table
        The table name to check

    where_sql
        The sql to select the row to check

    where_args
        The list parameters to substitute in where_sql
    """
    changes = {'name': name,
               'changes': {},
               'result': None,
               'comment': ''}
    conn = None
    try:
        conn = sqlite3.connect(db, detect_types=sqlite3.PARSE_DECLTYPES)
        conn.row_factory = _dict_factory
        rows = None
        if where_args is None:
            rows = _query(conn,
                          "SELECT * FROM `" + table + "` WHERE " + where_sql)
        else:
            rows = _query(conn,
                          "SELECT * FROM `" + table + "` WHERE " + where_sql,
                          where_args)

        if len(rows) > 1:
            changes['result'] = False
            changes['comment'] = "More than one row matched the specified query"
        elif len(rows) == 1:
            if __opts__['test']:
                changes['result'] = True
                changes['comment'] = "Row will be removed in " + table
                changes['changes']['old'] = rows[0]

            else:
                cursor = conn.execute("DELETE FROM `" +
                                      table + "` WHERE " + where_sql,
                                      where_args)
                conn.commit()
                if cursor.rowcount == 1:
                    changes['result'] = True
                    changes['comment'] = "Row removed"
                    changes['changes']['old'] = rows[0]
                else:
                    changes['result'] = False
                    changes['comment'] = "Unable to remove row"
        else:
            changes['result'] = True
            changes['comment'] = 'Row is absent'

    except Exception as e:
        changes['result'] = False
        changes['comment'] = str(e)

    finally:
        if conn:
            conn.close()

    return changes


def row_present(name,
                db,
                table,
                data,
                where_sql,
                where_args=None,
                update=False):
    """
    Checks to make sure the given row exists. If row exists and update is True
    then row will be updated with data. Otherwise it will leave existing
    row unmodified and check it against data. If the existing data
    doesn't match data_check the state will fail.  If the row doesn't
    exist then it will insert data into the table. If more than one
    row matches, then the state will fail.

    name
        Only used as the unique ID

    db
        The database file name

    table
        The table name to check the data

    data
        The dictionary of key/value pairs to check against if
        row exists, insert into the table if it doesn't

    where_sql
        The sql to select the row to check

    where_args
        The list parameters to substitute in where_sql

    update
        True will replace the existing row with data
        When False and the row exists and data does not equal
        the row data then the state will fail
    """
    changes = {'name': name,
               'changes': {},
               'result': None,
               'comment': ''}
    conn = None
    try:
        conn = sqlite3.connect(db, detect_types=sqlite3.PARSE_DECLTYPES)
        conn.row_factory = _dict_factory
        rows = None
        if where_args is None:
            rows = _query(conn, "SELECT * FROM `" +
                                table + "` WHERE " + where_sql)
        else:
            rows = _query(conn, "SELECT * FROM `" +
                                table + "` WHERE " + where_sql,
                                where_args)

        if len(rows) > 1:
            changes['result'] = False
            changes['comment'] = 'More than one row matched the specified query'
        elif len(rows) == 1:
            for key, value in six.iteritems(data):
                if key in rows[0] and rows[0][key] != value:
                    if update:
                        if __opts__['test']:
                            changes['result'] = True
                            changes['comment'] = "Row will be update in " + table

                        else:
                            columns = []
                            params = []
                            for key, value in six.iteritems(data):
                                columns.append("`" + key + "`=?")
                                params.append(value)

                            if where_args is not None:
                                params += where_args

                            sql = "UPDATE `" + table + "` SET "
                            sql += ",".join(columns)
                            sql += " WHERE "
                            sql += where_sql
                            cursor = conn.execute(sql, params)
                            conn.commit()
                            if cursor.rowcount == 1:
                                changes['result'] = True
                                changes['comment'] = "Row updated"
                                changes['changes']['old'] = rows[0]
                                changes['changes']['new'] = data
                            else:
                                changes['result'] = False
                                changes['comment'] = "Row update failed"
                    else:
                        changes['result'] = False
                        changes['comment'] = "Existing data does" + \
                                             "not match desired state"
                        break

            if changes['result'] is None:
                changes['result'] = True
                changes['comment'] = "Row exists"
        else:
            if __opts__['test']:
                changes['result'] = True
                changes['changes']['new'] = data
                changes['comment'] = "Row will be inserted into " + table
            else:
                columns = []
                value_stmt = []
                values = []
                for key, value in six.iteritems(data):
                    value_stmt.append('?')
                    values.append(value)
                    columns.append("`" + key + "`")

                sql = "INSERT INTO `" + table + "` ("
                sql += ",".join(columns)
                sql += ") VALUES ("
                sql += ",".join(value_stmt)
                sql += ")"
                cursor = conn.execute(sql, values)
                conn.commit()
                if cursor.rowcount == 1:
                    changes['result'] = True
                    changes['changes']['new'] = data
                    changes['comment'] = 'Inserted row'
                else:
                    changes['result'] = False
                    changes['comment'] = "Unable to insert data"

    except Exception as e:
        changes['result'] = False
        changes['comment'] = str(e)

    finally:
        if conn:
            conn.close()

    return changes


def table_absent(name, db):
    """
    Make sure the specified table does not exist

    name
        The name of the table

    db
        The name of the database file
    """
    changes = {'name': name,
               'changes': {},
               'result': None,
               'comment': ''}
    conn = None
    try:
        conn = sqlite3.connect(db, detect_types=sqlite3.PARSE_DECLTYPES)
        tables = _query(conn, "SELECT sql FROM sqlite_master " +
                              " WHERE type='table' AND name=?", [name])

        if len(tables) == 1:
            if __opts__['test']:
                changes['result'] = True
                changes['comment'] = "'" + name + "' will be dropped"
            else:
                conn.execute("DROP TABLE " + name)
                conn.commit()
                changes['changes']['old'] = tables[0][0]
                changes['result'] = True
                changes['comment'] = "'" + name + "' was dropped"
        elif len(tables) == 0:
            changes['result'] = True
            changes['comment'] = "'" + name + "' is already absent"
        else:
            changes['result'] = False
            changes['comment'] = "Multiple tables with the same name='" + \
                                 name + "'"

    except Exception as e:
        changes['result'] = False
        changes['comment'] = str(e)

    finally:
        if conn:
            conn.close()

    return changes


def table_present(name, db, schema, force=False):
    """
    Make sure the specified table exists with the specified schema

    name
        The name of the table

    db
        The name of the database file

    schema
        The dictionary containing the schema information

    force
        If the name of the table exists and force is set to False,
        the state will fail.  If force is set to True, the existing
        table will be replaced with the new table
    """
    changes = {'name': name,
               'changes': {},
               'result': None,
               'comment': ''}
    conn = None
    try:
        conn = sqlite3.connect(db, detect_types=sqlite3.PARSE_DECLTYPES)
        tables = _query(conn,
                        "SELECT sql FROM sqlite_master " +
                        "WHERE type='table' AND name=?", [name])

        if len(tables) == 1:
            sql = None
            if isinstance(schema, str):
                sql = schema
            else:
                sql = _get_sql_from_schema(name, schema)

            if sql != tables[0][0]:
                if force:
                    if __opts__['test']:
                        changes['result'] = True
                        changes['changes']['old'] = tables[0][0]
                        changes['changes']['new'] = sql
                        changes['comment'] = "'" + name + "' will be replaced"
                    else:
                        conn.execute("DROP TABLE `" + name + "`")
                        conn.execute(sql)
                        conn.commit()
                        changes['result'] = True
                        changes['changes']['old'] = tables[0][0]
                        changes['changes']['new'] = sql
                        changes['comment'] = "Replaced '" + name + "'"
                else:
                    changes['result'] = False
                    changes['comment'] = "Expected schema=" + sql + \
                                         "\nactual schema=" + tables[0][0]
            else:
                changes['result'] = True
                changes['comment'] = "'" + name + \
                                     "' exists with matching schema"
        elif len(tables) == 0:
            # Create the table
            sql = None
            if isinstance(schema, str):
                sql = schema
            else:
                sql = _get_sql_from_schema(name, schema)
            if __opts__['test']:
                changes['result'] = True
                changes['changes']['new'] = sql
                changes['comment'] = "'" + name + "' will be created"
            else:
                conn.execute(sql)
                conn.commit()
                changes['result'] = True
                changes['changes']['new'] = sql
                changes['comment'] = "Created table '" + name + "'"
        else:
            changes['result'] = False
            changes['comment'] = 'Multiple tables with the same name=' + name

    except Exception as e:
        changes['result'] = False
        changes['comment'] = str(e)

    finally:
        if conn:
            conn.close()

    return changes


def _query(conn, sql, parameters=None):
    cursor = None
    if parameters is None:
        cursor = conn.execute(sql)
    else:
        cursor = conn.execute(sql, parameters)
    return cursor.fetchall()


def _get_sql_from_schema(name, schema):
    return "CREATE TABLE `" + name + "` (" + ",".join(schema) + ")"


def _dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d



# -*- coding: utf-8 -*-

# Import Python libs
from __future__ import absolute_import
from distutils.version import LooseVersion  # pylint: disable=import-error,no-name-in-module
import logging
import random
import string

# Import Salt Testing libs
from salttesting.unit import skipIf, TestCase
from salttesting.mock import NO_MOCK, NO_MOCK_REASON, patch
from salttesting.helpers import ensure_in_syspath

ensure_in_syspath('../../')

# Import Salt libs
import salt.config
import salt.loader
from salt.modules import boto_cognitoidentity

# Import Mock libraries
from salttesting.mock import NO_MOCK, NO_MOCK_REASON, MagicMock, patch

# Import 3rd-party libs

# pylint: disable=import-error,no-name-in-module
try:
    import boto3
    from botocore.exceptions import ClientError
    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

from salt.ext.six.moves import range  # pylint: disable=import-error

# pylint: enable=import-error,no-name-in-module

# the boto_lambda module relies on the connect_to_region() method
# which was added in boto 2.8.0
# https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
required_boto3_version = '1.2.1'

region = 'us-east-1'
access_key = 'GKTADJGHEIQSXMKKRBJ08H'
secret_key = 'askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs'
conn_parameters = {'region': region, 'key': access_key, 'keyid': secret_key, 'profile': {}}
error_message = 'An error occurred (101) when calling the {0} operation: Test-defined error'
error_content = {
  'Error': {
    'Code': 101,
    'Message': "Test-defined error"
  }
}

first_pool_id = 'first_pool_id'
first_pool_name = 'first_pool'
second_pool_id = 'second_pool_id'
second_pool_name = 'second_pool'
second_pool_name_updated = 'second_pool_updated'
third_pool_id = 'third_pool_id'
third_pool_name = first_pool_name

identity_pools_ret = dict(IdentityPools=[dict(IdentityPoolId=first_pool_id,
                                              IdentityPoolName=first_pool_name),
                                         dict(IdentityPoolId=second_pool_id,
                                              IdentityPoolName=second_pool_name),
                                         dict(IdentityPoolId=third_pool_id,
                                              IdentityPoolName=third_pool_name)])

first_pool_ret = dict(IdentityPoolId=first_pool_id,
                      IdentityPoolName=first_pool_name,
                      AllowUnauthenticatedIdentities=False,
                      SupportedLoginProviders={'accounts.google.com': 'testing123',
                                               'api.twitter.com': 'testing123',
                                               'graph.facebook.com': 'testing123',
                                               'www.amazon.com': 'testing123'},
                      DeveloperProviderName='test_provider',
                      OpenIdConnectProviderARNs=['some_provider_arn',
                                                 'another_provider_arn'])

first_pool_role_ret = dict(IdentityPoolId=first_pool_id,
                           Roles=dict(authenticated='first_pool_auth_role',
                                      unauthenticated='first_pool_unauth_role'))

second_pool_ret = dict(IdentityPoolId=second_pool_id,
                       IdentityPoolName=second_pool_name,
                       AllowUnauthenticatedIdentities=False)

third_pool_ret = dict(IdentityPoolId=third_pool_id,
                      IdentityPoolName=third_pool_name,
                      AllowUnauthenticatedIdentities=False,
                      DeveloperProviderName='test_provider2')

third_pool_role_ret = dict(IdentityPoolId=third_pool_id)

default_pool_ret = dict(IdentityPoolId='default_pool_id',
                        IdentityPoolName='default_pool_name',
                        AllowUnauthenticatedIdentities=False,
                        DeveloperProviderName='test_provider_default')

default_pool_role_ret = dict(IdentityPoolId='default_pool_id')

log = logging.getLogger(__name__)

opts = salt.config.DEFAULT_MINION_OPTS
context = {}
utils = salt.loader.utils(opts, whitelist=['boto3'], context=context)

boto_cognitoidentity.__utils__ = utils
boto_cognitoidentity.__init__(opts)
boto_cognitoidentity.__salt__ = {}


def _has_required_boto():
    '''
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    '''
    if not HAS_BOTO:
        return False
    elif LooseVersion(boto3.__version__) < LooseVersion(required_boto3_version):
        return False
    else:
        return True


class BotoCognitoIdentityTestCaseBase(TestCase):
    conn = None

    # Set up MagicMock to replace the boto3 session
    def setUp(self):
        boto_cognitoidentity.__context__ = {}
        context.clear()
        # connections keep getting cached from prior tests, can't find the
        # correct context object to clear it. So randomize the cache key, to prevent any
        # cache hits
        conn_parameters['key'] = ''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(50))

        self.patcher = patch('boto3.session.Session')
        self.addCleanup(self.patcher.stop)
        mock_session = self.patcher.start()

        session_instance = mock_session.return_value
        self.conn = MagicMock()
        session_instance.client.return_value = self.conn


class BotoCognitoIdentityTestCaseMixin(object):
    pass


@skipIf(True, 'Skip these tests while investigating failures')
@skipIf(HAS_BOTO is False, 'The boto module must be installed.')
@skipIf(_has_required_boto() is False, 'The boto3 module must be greater than'
                                       ' or equal to version {0}'
        .format(required_boto3_version))
@skipIf(NO_MOCK, NO_MOCK_REASON)
class BotoCognitoIdentityTestCase(BotoCognitoIdentityTestCaseBase, BotoCognitoIdentityTestCaseMixin):
    '''
    TestCase for salt.modules.boto_cognitoidentity module
    '''

    def _describe_identity_pool_side_effect(self, *args, **kwargs):
        if kwargs.get('IdentityPoolId') == first_pool_id:
            return first_pool_ret
        elif kwargs.get('IdentityPoolId') == third_pool_id:
            return third_pool_ret
        else:
            return default_pool_ret

    def test_that_when_describing_a_named_identity_pool_and_pool_exists_the_describe_identity_pool_method_returns_pools_properties(self):
        '''
        Tests describing identity pool when the pool's name exists
        '''
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.side_effect = self._describe_identity_pool_side_effect
        result = boto_cognitoidentity.describe_identity_pools(IdentityPoolName=first_pool_name, **conn_parameters)
        self.assertEqual(result.get('identity_pools'), [first_pool_ret, third_pool_ret])

    def test_that_when_describing_a_identity_pool_by_its_id_and_pool_exists_the_desribe_identity_pool_method_returns_pools_properties(self):
        '''
        Tests describing identity pool when the given pool's id exists
        '''
        self.conn.describe_identity_pool.return_value = third_pool_ret
        result = boto_cognitoidentity.describe_identity_pools(IdentityPoolName='', IdentityPoolId=third_pool_id, **conn_parameters)
        self.assertEqual(result.get('identity_pools'), [third_pool_ret])
        self.conn.list_identity_pools.assert_not_called()

    def test_that_when_describing_a_named_identity_pool_and_pool_does_not_exist_the_describe_identity_pool_method_returns_none(self):
        '''
        Tests describing identity pool when the pool's name doesn't exist
        '''
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.return_value = first_pool_ret
        result = boto_cognitoidentity.describe_identity_pools(IdentityPoolName='no_such_pool', **conn_parameters)
        self.assertEqual(result.get('identity_pools', 'no such key'), None)

    def test_that_when_describing_a_named_identity_pool_and_error_thrown_the_describe_identity_pool_method_returns_error(self):
        '''
        Tests describing identity pool returns error when there is an exception to boto3 calls
        '''
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.side_effect = ClientError(error_content, 'error on describe identity pool')
        result = boto_cognitoidentity.describe_identity_pools(IdentityPoolName=first_pool_name, **conn_parameters)
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('error on describe identity pool'))

    def test_that_when_create_identity_pool_the_create_identity_pool_method_returns_created_identity_pool(self):
        '''
        Tests the positive case where create identity pool succeeds
        '''
        return_val = default_pool_ret.copy()
        return_val.pop('DeveloperProviderName', None)
        self.conn.create_identity_pool.return_value = return_val
        result = boto_cognitoidentity.create_identity_pool(IdentityPoolName='default_pool_name', **conn_parameters)
        mock_calls = self.conn.mock_calls
        #name, args, kwargs = mock_calls[0]
        self.assertTrue(result.get('created'))
        self.assertEqual(len(mock_calls), 1)
        self.assertEqual(mock_calls[0][0], 'create_identity_pool')
        self.assertNotIn('DeveloperProviderName', mock_calls[0][2])

    def test_that_when_create_identity_pool_and_error_thrown_the_create_identity_pool_method_returns_error(self):
        '''
        Tests the negative case where create identity pool has a boto3 client error
        '''
        self.conn.create_identity_pool.side_effect = ClientError(error_content, 'create_identity_pool')
        result = boto_cognitoidentity.create_identity_pool(IdentityPoolName='default_pool_name', **conn_parameters)
        self.assertIs(result.get('created'), False)
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('create_identity_pool'))

    def test_that_when_delete_identity_pools_with_multiple_matching_pool_names_the_delete_identity_pools_methos_returns_true_and_deleted_count(self):
        '''
        Tests that given 2 matching pool ids, the operation returns deleted status of true and
        count 2
        '''
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.delete_identity_pool.return_value = None
        result = boto_cognitoidentity.delete_identity_pools(IdentityPoolName=first_pool_name, **conn_parameters)
        mock_calls = self.conn.mock_calls
        self.assertTrue(result.get('deleted'))
        self.assertEqual(result.get('count'), 2)
        self.assertEqual(len(mock_calls), 3)
        self.assertEqual(mock_calls[1][0], 'delete_identity_pool')
        self.assertEqual(mock_calls[2][0], 'delete_identity_pool')
        self.assertEqual(mock_calls[1][2].get('IdentityPoolId'), first_pool_id)
        self.assertEqual(mock_calls[2][2].get('IdentityPoolId'), third_pool_id)

    def test_that_when_delete_identity_pools_with_no_matching_pool_names_the_delete_identity_pools_method_returns_false(self):
        '''
        Tests that the given pool name does not exist, the operation returns deleted status of false
        and count 0
        '''
        self.conn.list_identity_pools.return_value = identity_pools_ret
        result = boto_cognitoidentity.delete_identity_pools(IdentityPoolName='no_such_pool_name', **conn_parameters)
        mock_calls = self.conn.mock_calls
        self.assertIs(result.get('deleted'), False)
        self.assertEqual(result.get('count'), 0)
        self.assertEqual(len(mock_calls), 1)

    def test_that_when_delete_identity_pools_and_error_thrown_the_delete_identity_pools_method_returns_false_and_the_error(self):
        '''
        Tests that the delete_identity_pool method throws an exception
        '''
        self.conn.delete_identity_pool.side_effect = ClientError(error_content, 'delete_identity_pool')
        # try to delete an unexistent pool id "no_such_pool_id"
        result = boto_cognitoidentity.delete_identity_pools(IdentityPoolName=first_pool_name, IdentityPoolId='no_such_pool_id', **conn_parameters)
        mock_calls = self.conn.mock_calls
        self.assertIs(result.get('deleted'), False)
        self.assertIs(result.get('count'), None)
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('delete_identity_pool'))
        self.assertEqual(len(mock_calls), 1)

    def _get_identity_pool_roles_side_effect(self, *args, **kwargs):
        if kwargs.get('IdentityPoolId') == first_pool_id:
            return first_pool_role_ret
        elif kwargs.get('IdentityPoolId') == third_pool_id:
            return third_pool_role_ret
        else:
            return default_pool_role_ret

    def test_that_when_get_identity_pool_roles_with_matching_pool_names_the_get_identity_pool_roles_method_returns_pools_role_properties(self):
        '''
        Tests that the given 2 pool id's matching the given pool name, the results are
        passed through as a list of identity_pool_roles
        '''
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.get_identity_pool_roles.side_effect = self._get_identity_pool_roles_side_effect
        result = boto_cognitoidentity.get_identity_pool_roles(IdentityPoolName=first_pool_name, **conn_parameters)
        id_pool_roles = result.get('identity_pool_roles')
        self.assertIsNot(id_pool_roles, None)
        self.assertEqual(result.get('error', 'key_should_not_be_there'), 'key_should_not_be_there')
        self.assertEqual(len(id_pool_roles), 2)
        self.assertEqual(id_pool_roles[0], first_pool_role_ret)
        self.assertEqual(id_pool_roles[1], third_pool_role_ret)

    def test_that_when_get_identity_pool_roles_with_no_matching_pool_names_the_get_identity_pool_roles_method_returns_none(self):
        '''
        Tests that the given no pool id's matching the given pool name, the results returned is
        None and get_identity_pool_roles should never be called
        '''
        self.conn.list_identity_pools.return_value = identity_pools_ret
        result = boto_cognitoidentity.get_identity_pool_roles(IdentityPoolName="no_such_pool_name", **conn_parameters)
        mock_calls = self.conn.mock_calls
        self.assertIs(result.get('identity_pool_roles', 'key_should_be_there'), None)
        self.assertEqual(len(mock_calls), 1)
        self.assertEqual(result.get('error', 'key_should_not_be_there'), 'key_should_not_be_there')

    def test_that_when_get_identity_pool_roles_and_error_thrown_due_to_invalid_pool_id_the_get_identity_pool_roles_method_returns_error(self):
        '''
        Tests that given an invalid pool id, we properly handle error generated from get_identity_pool_roles
        '''
        self.conn.get_identity_pool_roles.side_effect = ClientError(error_content, 'get_identity_pool_roles')
        # try to delete an unexistent pool id "no_such_pool_id"
        result = boto_cognitoidentity.get_identity_pool_roles(IdentityPoolName='', IdentityPoolId='no_such_pool_id', **conn_parameters)
        mock_calls = self.conn.mock_calls
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('get_identity_pool_roles'))
        self.assertEqual(len(mock_calls), 1)

    def test_that_when_set_identity_pool_roles_with_invalid_pool_id_the_set_identity_pool_roles_method_returns_set_false_and_error(self):
        '''
        Tests that given an invalid pool id, we properly handle error generated from set_identity_pool_roles
        '''
        self.conn.set_identity_pool_roles.side_effect = ClientError(error_content, 'set_identity_pool_roles')
        result = boto_cognitoidentity.set_identity_pool_roles(IdentityPoolId='no_such_pool_id', **conn_parameters)
        mock_calls = self.conn.mock_calls
        self.assertIs(result.get('set'), False)
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('set_identity_pool_roles'))
        self.assertEqual(len(mock_calls), 1)

    def test_that_when_set_identity_pool_roles_with_no_roles_specified_the_set_identity_pool_roles_method_unset_the_roles(self):
        '''
        Tests that given a valid pool id, and no other roles given, the role for the pool is cleared.
        '''
        self.conn.set_identity_pool_roles.return_value = None
        result = boto_cognitoidentity.set_identity_pool_roles(IdentityPoolId='some_id', **conn_parameters)
        mock_calls = self.conn.mock_calls
        self.assertTrue(result.get('set'))
        self.assertEqual(len(mock_calls), 1)
        self.assertEqual(mock_calls[0][2].get('Roles'), {})

    def test_that_when_set_identity_pool_roles_with_only_auth_role_specified_the_set_identity_pool_roles_method_only_set_the_auth_role(self):
        '''
        Tests that given a valid pool id, and only other given role is the AuthenticatedRole, the auth role for the
        pool is set and unauth role is cleared.
        '''
        self.conn.set_identity_pool_roles.return_value = None
        with patch.dict(boto_cognitoidentity.__salt__, {'boto_iam.describe_role': MagicMock(return_value={'arn': 'my_auth_role_arn'})}):
            expected_roles = dict(authenticated='my_auth_role_arn')
            result = boto_cognitoidentity.set_identity_pool_roles(IdentityPoolId='some_id', AuthenticatedRole='my_auth_role', **conn_parameters)
            mock_calls = self.conn.mock_calls
            self.assertTrue(result.get('set'))
            self.assertEqual(len(mock_calls), 1)
            self.assertEqual(mock_calls[0][2].get('Roles'), expected_roles)

    def test_that_when_set_identity_pool_roles_with_only_unauth_role_specified_the_set_identity_pool_roles_method_only_set_the_unauth_role(self):
        '''
        Tests that given a valid pool id, and only other given role is the UnauthenticatedRole, the unauth role for the
        pool is set and the auth role is cleared.
        '''
        self.conn.set_identity_pool_roles.return_value = None
        with patch.dict(boto_cognitoidentity.__salt__, {'boto_iam.describe_role': MagicMock(return_value={'arn': 'my_unauth_role_arn'})}):
            expected_roles = dict(unauthenticated='my_unauth_role_arn')
            result = boto_cognitoidentity.set_identity_pool_roles(IdentityPoolId='some_id', UnauthenticatedRole='my_unauth_role', **conn_parameters)
            mock_calls = self.conn.mock_calls
            self.assertTrue(result.get('set'))
            self.assertEqual(len(mock_calls), 1)
            self.assertEqual(mock_calls[0][2].get('Roles'), expected_roles)

    def test_that_when_set_identity_pool_roles_with_both_roles_specified_the_set_identity_pool_role_method_set_both_roles(self):
        '''
        Tests setting of both roles to valid given roles
        '''
        self.conn.set_identity_pool_roles.return_value = None
        with patch.dict(boto_cognitoidentity.__salt__, {'boto_iam.describe_role': MagicMock(return_value={'arn': 'my_unauth_role_arn'})}):
            expected_roles = dict(authenticated='arn:aws:iam:my_auth_role',
                                  unauthenticated='my_unauth_role_arn')
            result = boto_cognitoidentity.set_identity_pool_roles(IdentityPoolId='some_id', AuthenticatedRole='arn:aws:iam:my_auth_role', UnauthenticatedRole='my_unauth_role', **conn_parameters)
            mock_calls = self.conn.mock_calls
            self.assertTrue(result.get('set'))
            self.assertEqual(len(mock_calls), 1)
            self.assertEqual(mock_calls[0][2].get('Roles'), expected_roles)

    def test_that_when_set_identity_pool_roles_given_invalid_auth_role_the_set_identity_pool_method_returns_set_false_and_error(self):
        '''
        Tests error handling for invalid auth role
        '''
        with patch.dict(boto_cognitoidentity.__salt__, {'boto_iam.describe_role': MagicMock(return_value=False)}):
            result = boto_cognitoidentity.set_identity_pool_roles(IdentityPoolId='some_id', AuthenticatedRole='no_such_auth_role', **conn_parameters)
            mock_calls = self.conn.mock_calls
            self.assertIs(result.get('set'), False)
            self.assertIn('no_such_auth_role', result.get('error', ''))
            self.assertEqual(len(mock_calls), 0)

    def test_that_when_set_identity_pool_roles_given_invalid_unauth_role_the_set_identity_pool_method_returns_set_false_and_error(self):
        '''
        Tests error handling for invalid unauth role
        '''
        with patch.dict(boto_cognitoidentity.__salt__, {'boto_iam.describe_role': MagicMock(return_value=False)}):
            result = boto_cognitoidentity.set_identity_pool_roles(IdentityPoolId='some_id', AuthenticatedRole='arn:aws:iam:my_auth_role', UnauthenticatedRole='no_such_unauth_role', **conn_parameters)
            mock_calls = self.conn.mock_calls
            self.assertIs(result.get('set'), False)
            self.assertIn('no_such_unauth_role', result.get('error', ''))
            self.assertEqual(len(mock_calls), 0)

    def test_that_when_update_identity_pool_given_invalid_pool_id_the_update_identity_pool_method_returns_updated_false_and_error(self):
        '''
        Tests error handling for invalid pool id
        '''
        self.conn.describe_identity_pool.side_effect = ClientError(error_content, 'error on describe identity pool with pool id')
        result = boto_cognitoidentity.update_identity_pool(IdentityPoolId='no_such_pool_id', **conn_parameters)
        self.assertIs(result.get('updated'), False)
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('error on describe identity pool with pool id'))

    def test_that_when_update_identity_pool_given_only_valid_pool_id_the_update_identity_pool_method_returns_udpated_identity(self):
        '''
        Tests the base case of calling update_identity_pool with only the pool id, verify
        that the passed parameters into boto3 update_identity_pool has at least all the
        expected required parameters in IdentityPoolId, IdentityPoolName and AllowUnauthenticatedIdentities
        '''
        self.conn.describe_identity_pool.return_value = second_pool_ret
        self.conn.update_identity_pool.return_value = second_pool_ret
        expected_params = second_pool_ret
        result = boto_cognitoidentity.update_identity_pool(IdentityPoolId=second_pool_id, **conn_parameters)
        self.assertTrue(result.get('updated'))
        self.assertEqual(result.get('identity_pool'), second_pool_ret)
        self.conn.update_identity_pool.assert_called_with(**expected_params)

    def test_that_when_update_identity_pool_given_valid_pool_id_and_pool_name_the_update_identity_pool_method_returns_updated_identity_pool(self):
        '''
        Tests successful update of a pool's name
        '''
        self.conn.describe_identity_pool.return_value = second_pool_ret
        second_pool_updated_ret = second_pool_ret.copy()
        second_pool_updated_ret['IdentityPoolName'] = second_pool_name_updated
        self.conn.update_identity_pool.return_value = second_pool_updated_ret
        expected_params = second_pool_updated_ret
        result = boto_cognitoidentity.update_identity_pool(IdentityPoolId=second_pool_id, IdentityPoolName=second_pool_name_updated, **conn_parameters)
        self.assertTrue(result.get('updated'))
        self.assertEqual(result.get('identity_pool'), second_pool_updated_ret)
        self.conn.update_identity_pool.assert_called_with(**expected_params)

    def test_that_when_update_identity_pool_given_empty_dictionary_for_supported_login_providers_the_update_identity_pool_method_is_called_with_proper_request_params(self):
        '''
        Tests the request parameters to boto3 update_identity_pool's AllowUnauthenticatedIdentities is {}
        '''
        self.conn.describe_identity_pool.return_value = first_pool_ret
        first_pool_updated_ret = first_pool_ret.copy()
        first_pool_updated_ret.pop('SupportedLoginProviders')
        self.conn.update_identity_pool.return_value = first_pool_updated_ret
        expected_params = first_pool_ret.copy()
        expected_params['SupportedLoginProviders'] = {}
        expected_params.pop('DeveloperProviderName')
        expected_params.pop('OpenIdConnectProviderARNs')
        result = boto_cognitoidentity.update_identity_pool(IdentityPoolId=first_pool_id, SupportedLoginProviders={}, **conn_parameters)
        self.assertTrue(result.get('updated'))
        self.assertEqual(result.get('identity_pool'), first_pool_updated_ret)
        self.conn.update_identity_pool.assert_called_with(**expected_params)

    def test_that_when_update_identity_pool_given_empty_list_for_openid_connect_provider_arns_the_update_identity_pool_method_is_called_with_proper_request_params(self):
        '''
        Tests the request parameters to boto3 update_identity_pool's OpenIdConnectProviderARNs is []
        '''
        self.conn.describe_identity_pool.return_value = first_pool_ret
        first_pool_updated_ret = first_pool_ret.copy()
        first_pool_updated_ret.pop('OpenIdConnectProviderARNs')
        self.conn.update_identity_pool.return_value = first_pool_updated_ret
        expected_params = first_pool_ret.copy()
        expected_params.pop('SupportedLoginProviders')
        expected_params.pop('DeveloperProviderName')
        expected_params['OpenIdConnectProviderARNs'] = []
        result = boto_cognitoidentity.update_identity_pool(IdentityPoolId=first_pool_id, OpenIdConnectProviderARNs=[], **conn_parameters)
        self.assertTrue(result.get('updated'))
        self.assertEqual(result.get('identity_pool'), first_pool_updated_ret)
        self.conn.update_identity_pool.assert_called_with(**expected_params)

    def test_that_when_update_identity_pool_given_developer_provider_name_when_developer_provider_name_was_set_previously_the_udpate_identity_pool_method_is_called_without_developer_provider_name_param(self):
        '''
        Tests the request parameters do not include 'DeveloperProviderName' if this was previously set
        for the given pool id
        '''
        self.conn.describe_identity_pool.return_value = first_pool_ret
        self.conn.update_identity_pool.return_value = first_pool_ret
        expected_params = first_pool_ret.copy()
        expected_params.pop('SupportedLoginProviders')
        expected_params.pop('DeveloperProviderName')
        expected_params.pop('OpenIdConnectProviderARNs')
        result = boto_cognitoidentity.update_identity_pool(IdentityPoolId=first_pool_id, DeveloperProviderName='this should not change', **conn_parameters)
        self.assertTrue(result.get('updated'))
        self.assertEqual(result.get('identity_pool'), first_pool_ret)
        self.conn.update_identity_pool.assert_called_with(**expected_params)

    def test_that_when_update_identity_pool_given_developer_provider_name_is_included_in_the_params_when_associated_for_the_first_time(self):
        '''
        Tests the request parameters include 'DeveloperProviderName' when the pool did not have this
        property set previously.
        '''
        self.conn.describe_identity_pool.return_value = second_pool_ret
        second_pool_updated_ret = second_pool_ret.copy()
        second_pool_updated_ret['DeveloperProviderName'] = 'added_developer_provider'
        self.conn.update_identity_pool.return_value = second_pool_updated_ret
        expected_params = second_pool_updated_ret.copy()
        result = boto_cognitoidentity.update_identity_pool(IdentityPoolId=second_pool_id, DeveloperProviderName='added_developer_provider', **conn_parameters)
        self.assertTrue(result.get('updated'))
        self.assertEqual(result.get('identity_pool'), second_pool_updated_ret)
        self.conn.update_identity_pool.assert_called_with(**expected_params)

    def test_that_the_update_identity_pool_method_handles_exception_from_boto3(self):
        '''
        Tests the error handling of exception generated by boto3 update_identity_pool
        '''
        self.conn.describe_identity_pool.return_value = second_pool_ret
        second_pool_updated_ret = second_pool_ret.copy()
        second_pool_updated_ret['DeveloperProviderName'] = 'added_developer_provider'
        self.conn.update_identity_pool.side_effect = ClientError(error_content, 'update_identity_pool')
        result = boto_cognitoidentity.update_identity_pool(IdentityPoolId=second_pool_id, DeveloperProviderName='added_developer_provider', **conn_parameters)
        self.assertIs(result.get('updated'), False)
        self.assertEqual(result.get('error', {}).get('message'), error_message.format('update_identity_pool'))

if __name__ == '__main__':
    from integration import run_tests  # pylint: disable=import-error
    run_tests(BotoCognitoIdentityTestCase, needs_daemon=False)



# -*- coding: utf-8 -*-
'''
Manage CloudTrail Objects
=========================

.. versionadded:: 2016.3.0

Create and destroy CloudTrail objects. Be aware that this interacts with Amazon's services,
and so may incur charges.

:depends:
    - boto
    - boto3

The dependencies listed above can be installed via package or pip.

This module accepts explicit vpc credentials but can also utilize
IAM roles assigned to the instance through Instance Profiles. Dynamic
credentials are then automatically obtained from AWS API and no further
configuration is necessary. More information available `here
<http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html>`_.

If IAM roles are not used you need to specify them either in a pillar file or
in the minion's config file:

.. code-block:: yaml

    vpc.keyid: GKTADJGHEIQSXMKKRBJ08H
    vpc.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

It's also possible to specify ``key``, ``keyid`` and ``region`` via a profile,
either passed in as a dict, or as a string to pull from pillars or minion
config:

.. code-block:: yaml

    myprofile:
        keyid: GKTADJGHEIQSXMKKRBJ08H
        key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-east-1

.. code-block:: yaml

    Ensure trail exists:
        boto_cloudtrail.present:
            - Name: mytrail
            - S3BucketName: mybucket
            - S3KeyPrefix: prefix
            - region: us-east-1
            - keyid: GKTADJGHEIQSXMKKRBJ08H
            - key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

'''

# Import Python Libs
from __future__ import absolute_import
import logging
import os
import os.path

# Import Salt Libs
import salt.ext.six as six
import salt.utils

log = logging.getLogger(__name__)


def __virtual__():
    '''
    Only load if boto is available.
    '''
    return 'boto_cloudtrail' if 'boto_cloudtrail.exists' in __salt__ else False


def present(name, Name,
           S3BucketName, S3KeyPrefix=None,
           SnsTopicName=None,
           IncludeGlobalServiceEvents=True,
           IsMultiRegionTrail=None,
           EnableLogFileValidation=False,
           CloudWatchLogsLogGroupArn=None,
           CloudWatchLogsRoleArn=None,
           KmsKeyId=None,
           LoggingEnabled=True,
           Tags=None,
           region=None, key=None, keyid=None, profile=None):
    '''
    Ensure trail exists.

    name
        The name of the state definition

    Name
        Name of the trail.

    S3BucketName
        Specifies the name of the Amazon S3 bucket designated for publishing log
        files.

    S3KeyPrefix
        Specifies the Amazon S3 key prefix that comes after the name of the
        bucket you have designated for log file delivery.

    SnsTopicName
        Specifies the name of the Amazon SNS topic defined for notification of
        log file delivery. The maximum length is 256 characters.

    IncludeGlobalServiceEvents
        Specifies whether the trail is publishing events from global services
        such as IAM to the log files.

    EnableLogFileValidation
        Specifies whether log file integrity validation is enabled. The default
        is false.

    CloudWatchLogsLogGroupArn
        Specifies a log group name using an Amazon Resource Name (ARN), a unique
        identifier that represents the log group to which CloudTrail logs will
        be delivered. Not required unless you specify CloudWatchLogsRoleArn.

    CloudWatchLogsRoleArn
        Specifies the role for the CloudWatch Logs endpoint to assume to write
        to a user's log group.

    KmsKeyId
        Specifies the KMS key ID to use to encrypt the logs delivered by
        CloudTrail. The value can be a an alias name prefixed by "alias/", a
        fully specified ARN to an alias, a fully specified ARN to a key, or a
        globally unique identifier.

    LoggingEnabled
        Whether logging should be enabled for the trail

    Tags
        A dictionary of tags that should be set on the trail

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string) that
        contains a dict with region, key and keyid.
    '''
    ret = {'name': Name,
           'result': True,
           'comment': '',
           'changes': {}
           }

    r = __salt__['boto_cloudtrail.exists'](Name=Name,
           region=region, key=key, keyid=keyid, profile=profile)

    if 'error' in r:
        ret['result'] = False
        ret['comment'] = 'Failed to create trail: {0}.'.format(r['error']['message'])
        return ret

    if not r.get('exists'):
        if __opts__['test']:
            ret['comment'] = 'CloudTrail {0} is set to be created.'.format(Name)
            ret['result'] = None
            return ret
        r = __salt__['boto_cloudtrail.create'](Name=Name,
                   S3BucketName=S3BucketName,
                   S3KeyPrefix=S3KeyPrefix,
                   SnsTopicName=SnsTopicName,
                   IncludeGlobalServiceEvents=IncludeGlobalServiceEvents,
                   IsMultiRegionTrail=IsMultiRegionTrail,
                   EnableLogFileValidation=EnableLogFileValidation,
                   CloudWatchLogsLogGroupArn=CloudWatchLogsLogGroupArn,
                   CloudWatchLogsRoleArn=CloudWatchLogsRoleArn,
                   KmsKeyId=KmsKeyId,
                   region=region, key=key, keyid=keyid, profile=profile)
        if not r.get('created'):
            ret['result'] = False
            ret['comment'] = 'Failed to create trail: {0}.'.format(r['error']['message'])
            return ret
        _describe = __salt__['boto_cloudtrail.describe'](Name,
                                   region=region, key=key, keyid=keyid, profile=profile)
        ret['changes']['old'] = {'trail': None}
        ret['changes']['new'] = _describe
        ret['comment'] = 'CloudTrail {0} created.'.format(Name)

        if LoggingEnabled:
            r = __salt__['boto_cloudtrail.start_logging'](Name=Name,
                   region=region, key=key, keyid=keyid, profile=profile)
            if 'error' in r:
                ret['result'] = False
                ret['comment'] = 'Failed to create trail: {0}.'.format(r['error']['message'])
                ret['changes'] = {}
                return ret
            ret['changes']['new']['trail']['LoggingEnabled'] = True
        else:
            ret['changes']['new']['trail']['LoggingEnabled'] = False

        if bool(Tags):
            r = __salt__['boto_cloudtrail.add_tags'](Name=Name,
                   region=region, key=key, keyid=keyid, profile=profile, **Tags)
            if not r.get('tagged'):
                ret['result'] = False
                ret['comment'] = 'Failed to create trail: {0}.'.format(r['error']['message'])
                ret['changes'] = {}
                return ret
            ret['changes']['new']['trail']['Tags'] = Tags
        return ret

    ret['comment'] = os.linesep.join([ret['comment'], 'CloudTrail {0} is present.'.format(Name)])
    ret['changes'] = {}
    # trail exists, ensure config matches
    _describe = __salt__['boto_cloudtrail.describe'](Name=Name,
                                  region=region, key=key, keyid=keyid, profile=profile)
    if 'error' in _describe:
        ret['result'] = False
        ret['comment'] = 'Failed to update trail: {0}.'.format(_describe['error']['message'])
        ret['changes'] = {}
        return ret
    _describe = _describe.get('trail')

    r = __salt__['boto_cloudtrail.status'](Name=Name,
                   region=region, key=key, keyid=keyid, profile=profile)
    _describe['LoggingEnabled'] = r.get('trail', {}).get('IsLogging', False)

    need_update = False
    bucket_vars = {'S3BucketName': 'S3BucketName',
                   'S3KeyPrefix': 'S3KeyPrefix',
                   'SnsTopicName': 'SnsTopicName',
                   'IncludeGlobalServiceEvents': 'IncludeGlobalServiceEvents',
                   'IsMultiRegionTrail': 'IsMultiRegionTrail',
                   'EnableLogFileValidation': 'LogFileValidationEnabled',
                   'CloudWatchLogsLogGroupArn': 'CloudWatchLogsLogGroupArn',
                   'CloudWatchLogsRoleArn': 'CloudWatchLogsRoleArn',
                   'KmsKeyId': 'KmsKeyId',
                   'LoggingEnabled': 'LoggingEnabled'}

    for invar, outvar in six.iteritems(bucket_vars):
        if _describe[outvar] != locals()[invar]:
            need_update = True
            ret['changes'].setdefault('new', {})[invar] = locals()[invar]
            ret['changes'].setdefault('old', {})[invar] = _describe[outvar]

    r = __salt__['boto_cloudtrail.list_tags'](Name=Name,
                   region=region, key=key, keyid=keyid, profile=profile)
    _describe['Tags'] = r.get('tags', {})
    tagchange = salt.utils.compare_dicts(_describe['Tags'], Tags)
    if bool(tagchange):
        need_update = True
        ret['changes'].setdefault('new', {})['Tags'] = Tags
        ret['changes'].setdefault('old', {})['Tags'] = _describe['Tags']

    if need_update:
        if __opts__['test']:
            msg = 'CloudTrail {0} set to be modified.'.format(Name)
            ret['comment'] = msg
            ret['result'] = None
            return ret

        ret['comment'] = os.linesep.join([ret['comment'], 'CloudTrail to be modified'])
        r = __salt__['boto_cloudtrail.update'](Name=Name,
                   S3BucketName=S3BucketName,
                   S3KeyPrefix=S3KeyPrefix,
                   SnsTopicName=SnsTopicName,
                   IncludeGlobalServiceEvents=IncludeGlobalServiceEvents,
                   IsMultiRegionTrail=IsMultiRegionTrail,
                   EnableLogFileValidation=EnableLogFileValidation,
                   CloudWatchLogsLogGroupArn=CloudWatchLogsLogGroupArn,
                   CloudWatchLogsRoleArn=CloudWatchLogsRoleArn,
                   KmsKeyId=KmsKeyId,
                   region=region, key=key, keyid=keyid, profile=profile)
        if not r.get('updated'):
            ret['result'] = False
            ret['comment'] = 'Failed to update trail: {0}.'.format(r['error']['message'])
            ret['changes'] = {}
            return ret

        if LoggingEnabled:
            r = __salt__['boto_cloudtrail.start_logging'](Name=Name,
                   region=region, key=key, keyid=keyid, profile=profile)
            if not r.get('started'):
                ret['result'] = False
                ret['comment'] = 'Failed to update trail: {0}.'.format(r['error']['message'])
                ret['changes'] = {}
                return ret
        else:
            r = __salt__['boto_cloudtrail.stop_logging'](Name=Name,
                   region=region, key=key, keyid=keyid, profile=profile)
            if not r.get('stopped'):
                ret['result'] = False
                ret['comment'] = 'Failed to update trail: {0}.'.format(r['error']['message'])
                ret['changes'] = {}
                return ret

        if bool(tagchange):
            adds = {}
            removes = {}
            for k, diff in six.iteritems(tagchange):
                if diff.get('new', '') != '':
                    # there's an update for this key
                    adds[k] = Tags[k]
                elif diff.get('old', '') != '':
                    removes[k] = _describe['Tags'][k]
            if bool(adds):
                r = __salt__['boto_cloudtrail.add_tags'](Name=Name,
                   region=region, key=key, keyid=keyid, profile=profile, **adds)
            if bool(removes):
                r = __salt__['boto_cloudtrail.remove_tags'](Name=Name,
                   region=region, key=key, keyid=keyid, profile=profile,
                   **removes)

    return ret


def absent(name, Name,
                  region=None, key=None, keyid=None, profile=None):
    '''
    Ensure trail with passed properties is absent.

    name
        The name of the state definition.

    Name
        Name of the trail.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string) that
        contains a dict with region, key and keyid.
    '''

    ret = {'name': Name,
           'result': True,
           'comment': '',
           'changes': {}
           }

    r = __salt__['boto_cloudtrail.exists'](Name,
                       region=region, key=key, keyid=keyid, profile=profile)
    if 'error' in r:
        ret['result'] = False
        ret['comment'] = 'Failed to delete trail: {0}.'.format(r['error']['message'])
        return ret

    if r and not r['exists']:
        ret['comment'] = 'CloudTrail {0} does not exist.'.format(Name)
        return ret

    if __opts__['test']:
        ret['comment'] = 'CloudTrail {0} is set to be removed.'.format(Name)
        ret['result'] = None
        return ret
    r = __salt__['boto_cloudtrail.delete'](Name,
                                    region=region, key=key,
                                    keyid=keyid, profile=profile)
    if not r['deleted']:
        ret['result'] = False
        ret['comment'] = 'Failed to delete trail: {0}.'.format(r['error']['message'])
        return ret
    ret['changes']['old'] = {'trail': Name}
    ret['changes']['new'] = {'trail': None}
    ret['comment'] = 'CloudTrail {0} deleted.'.format(Name)
    return ret



# -*- coding: utf-8 -*-
'''
Management of Open vSwitch ports.
'''


def __virtual__():
    '''
    Only make these states available if Open vSwitch module is available.
    '''
    return 'openvswitch.port_add' in __salt__


def present(name, bridge, type=None, id=None, remote=None, dst_port=None):
    '''
    Ensures that the named port exists on bridge, eventually creates it.

    Args:
        name: The name of the port.
        bridge: The name of the bridge.
        type: Optional type of interface to create, currently supports: vlan, vxlan and gre.
        id: Optional tunnel's key.
        remote: Remote endpoint's IP address.
        dst_port: Port to use when creating tunnelport in the switch.

    '''
    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}
    types = ('vlan', 'vxlan', 'gre')

    if type and type not in types:
        raise TypeError('The optional type argument must be one of these values: {0}.'.format(str(types)))

    bridge_exists = __salt__['openvswitch.bridge_exists'](bridge)

    if bridge_exists:
        port_list = __salt__['openvswitch.port_list'](bridge)

    # Comment and change messages
    comment_bridge_notexists = 'Bridge {0} does not exist.'.format(bridge)
    comment_port_exists = 'Port {0} already exists.'.format(name)
    comment_port_created = 'Port {0} created on bridge {1}.'.format(name, bridge)
    comment_port_notcreated = 'Unable to create port {0} on bridge {1}.'.format(name, bridge)
    changes_port_created = {name: {'old': 'No port named {0} present.'.format(name),
                                   'new': 'Created port {1} on bridge {0}.'.format(bridge, name),
                                   }
                            }

    comment_vlan_invalid_id = 'VLANs id must be between 0 and 4095.'
    comment_vlan_invalid_name = 'Could not find network interface {0}.'.format(name)
    comment_vlan_port_exists = 'Port {0} with access to VLAN {1} already exists on bridge {2}.'.format(name, id, bridge)
    comment_vlan_created = 'Created port {0} with access to VLAN {1} on bridge {2}.'.format(name, id, bridge)
    comment_vlan_notcreated = 'Unable to create port {0} with access to VLAN {1} on ' \
                              'bridge {2}.'.format(name, id, bridge)
    changes_vlan_created = {name: {'old': 'No port named {0} with access to VLAN {1} present on '
                                          'bridge {2} present.'.format(name, id, bridge),
                                   'new': 'Created port {1} with access to VLAN {2} on '
                                          'bridge {0}.'.format(bridge, name, id),
                                   }
                            }

    comment_gre_invalid_id = 'Id of GRE tunnel must be an unsigned 32-bit integer.'
    comment_gre_interface_exists = 'GRE tunnel interface {0} with rempte ip {1} and key {2} ' \
                                   'already exists on bridge {3}.'.format(name, remote, id, bridge)
    comment_gre_created = 'Created GRE tunnel interface {0} with remote ip {1}  and key {2} ' \
                          'on bridge {3}.'.format(name, remote, id, bridge)
    comment_gre_notcreated = 'Unable to create GRE tunnel interface {0} with remote ip {1} and key {2} ' \
                             'on bridge {3}.'.format(name, remote, id, bridge)
    changes_gre_created = {name: {'old': 'No GRE tunnel interface {0} with remote ip {1} and key {2} '
                                         'on bridge {3} present.'.format(name, remote, id, bridge),
                                   'new': 'Created GRE tunnel interface {0} with remote ip {1} and key {2} '
                                          'on bridge {3}.'.format(name, remote, id, bridge),
                                   }
                            }

    comment_dstport = ' (dst_port' + str(dst_port) + ')' if 0 < dst_port <= 65535 else ''
    comment_vxlan_invalid_id = 'Id of VXLAN tunnel must be an unsigned 64-bit integer.'
    comment_vxlan_interface_exists = 'VXLAN tunnel interface {0} with rempte ip {1} and key {2} ' \
                                   'already exists on bridge {3}{4}.'.format(name, remote, id, bridge, comment_dstport)
    comment_vxlan_created = 'Created VXLAN tunnel interface {0} with remote ip {1}  and key {2} ' \
                          'on bridge {3}{4}.'.format(name, remote, id, bridge, comment_dstport)
    comment_vxlan_notcreated = 'Unable to create VXLAN tunnel interface {0} with remote ip {1} and key {2} ' \
                             'on bridge {3}{4}.'.format(name, remote, id, bridge, comment_dstport)
    changes_vxlan_created = {name: {'old': 'No VXLAN tunnel interface {0} with remote ip {1} and key {2} '
                                         'on bridge {3}{4} present.'.format(name, remote, id, bridge, comment_dstport),
                                   'new': 'Created VXLAN tunnel interface {0} with remote ip {1} and key {2} '
                                          'on bridge {3}{4}.'.format(name, remote, id, bridge, comment_dstport),
                                   }
                            }
    comment_invalid_ip = 'Remote is not valid ip address.'

    # Check VLANs attributes
    def _check_vlan():
        tag = __salt__['openvswitch.port_get_tag'](name)
        interfaces = __salt__['network.interfaces']()
        if not 0 <= id <= 4095:
            ret['result'] = False
            ret['comment'] = comment_vlan_invalid_id
        elif name not in interfaces:
            ret['result'] = False
            ret['comment'] = comment_vlan_invalid_name
        elif tag and name in port_list:
            try:
                if int(tag[0]) == id:
                    ret['result'] = True
                    ret['comment'] = comment_vlan_port_exists
            except (ValueError, KeyError):
                pass

    # Check GRE tunnels attributes
    def _check_gre():
        interface_options = __salt__['openvswitch.interface_get_options'](name)
        interface_type = __salt__['openvswitch.interface_get_type'](name)
        if not 0 <= id <= 2**32:
            ret['result'] = False
            ret['comment'] = comment_gre_invalid_id
        elif not __salt__['dig.check_ip'](remote):
            ret['result'] = False
            ret['comment'] = comment_invalid_ip
        elif interface_options and interface_type and name in port_list:
            interface_attroptions = '{key=\"' + str(id) + '\", remote_ip=\"' + str(remote) + '\"}'
            try:
                if interface_type[0] == 'gre' and interface_options[0] == interface_attroptions:
                    ret['result'] = True
                    ret['comment'] = comment_gre_interface_exists
            except KeyError:
                pass

    # Check VXLAN tunnels attributes
    def _check_vxlan():
        interface_options = __salt__['openvswitch.interface_get_options'](name)
        interface_type = __salt__['openvswitch.interface_get_type'](name)
        if not 0 <= id <= 2**64:
            ret['result'] = False
            ret['comment'] = comment_vxlan_invalid_id
        elif not __salt__['dig.check_ip'](remote):
            ret['result'] = False
            ret['comment'] = comment_invalid_ip
        elif interface_options and interface_type and name in port_list:
            opt_port = 'dst_port=\"' + str(dst_port) + '\", ' if 0 < dst_port <= 65535 else ''
            interface_attroptions = '{{{0}key=\"'.format(opt_port) + str(id) + '\", remote_ip=\"' + str(remote) + '\"}'
            try:
                if interface_type[0] == 'vxlan' and interface_options[0] == interface_attroptions:
                    ret['result'] = True
                    ret['comment'] = comment_vxlan_interface_exists
            except KeyError:
                pass

    # Dry run, test=true mode
    if __opts__['test']:
        if bridge_exists:
            if type == 'vlan':
                _check_vlan()
                if not ret['comment']:
                    ret['result'] = None
                    ret['comment'] = comment_vlan_created
            elif type == 'vxlan':
                _check_vxlan()
                if not ret['comment']:
                    ret['result'] = None
                    ret['comment'] = comment_vxlan_created
            elif type == 'gre':
                _check_gre()
                if not ret['comment']:
                    ret['result'] = None
                    ret['comment'] = comment_gre_created
            else:
                if name in port_list:
                    ret['result'] = True
                    ret['comment'] = comment_port_exists
                else:
                    ret['result'] = None
                    ret['comment'] = comment_port_created
        else:
            ret['result'] = None
            ret['comment'] = comment_bridge_notexists

        return ret

    if bridge_exists:
        if type == 'vlan':
            _check_vlan()
            if not ret['comment']:
                port_create_vlan = __salt__['openvswitch.port_create_vlan'](bridge, name, id)
                if port_create_vlan:
                    ret['result'] = True
                    ret['comment'] = comment_vlan_created
                    ret['changes'] = changes_vlan_created
                else:
                    ret['result'] = False
                    ret['comment'] = comment_vlan_notcreated
        elif type == 'vxlan':
            _check_vxlan()
            if not ret['comment']:
                port_create_vxlan = __salt__['openvswitch.port_create_vxlan'](bridge, name, id, remote, dst_port)
                if port_create_vxlan:
                    ret['result'] = True
                    ret['comment'] = comment_vxlan_created
                    ret['changes'] = changes_vxlan_created
                else:
                    ret['result'] = False
                    ret['comment'] = comment_vxlan_notcreated
        elif type == 'gre':
            _check_gre()
            if not ret['comment']:
                port_create_gre = __salt__['openvswitch.port_create_gre'](bridge, name, id, remote)
                if port_create_gre:
                    ret['result'] = True
                    ret['comment'] = comment_gre_created
                    ret['changes'] = changes_gre_created
                else:
                    ret['result'] = False
                    ret['comment'] = comment_gre_notcreated
        else:
            if name in port_list:
                ret['result'] = True
                ret['comment'] = comment_port_exists
            else:
                port_add = __salt__['openvswitch.port_add'](bridge, name)
                if port_add:
                    ret['result'] = True
                    ret['comment'] = comment_port_created
                    ret['changes'] = changes_port_created
                else:
                    ret['result'] = False
                    ret['comment'] = comment_port_notcreated
    else:
        ret['result'] = False
        ret['comment'] = comment_bridge_notexists

    return ret


def absent(name, bridge=None):
    '''
    Ensures that the named port exists on bridge, eventually deletes it.
    If bridge is not set, port is removed from  whatever bridge contains it.

    Args:
        name: The name of the port.
        bridge: The name of the bridge.

    '''
    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if bridge:
        bridge_exists = __salt__['openvswitch.bridge_exists'](bridge)
        if bridge_exists:
            port_list = __salt__['openvswitch.port_list'](bridge)
        else:
            port_list = ()
    else:
        port_list = [name]

    # Comment and change messages
    comment_bridge_notexists = 'Bridge {0} does not exist.'.format(bridge)
    comment_port_notexists = 'Port {0} does not exist on bridge {1}.'.format(name, bridge)
    comment_port_deleted = 'Port {0} deleted.'.format(name)
    comment_port_notdeleted = 'Unable to delete port {0}.'.format(name)
    changes_port_deleted = {name: {'old': 'Port named {0} may exist.'.format(name),
                                   'new': 'Deleted port {0}.'.format(name),
                                   }
                            }

    # Dry run, test=true mode
    if __opts__['test']:
        if bridge and not bridge_exists:
            ret['result'] = None
            ret['comment'] = comment_bridge_notexists
        elif name not in port_list:
            ret['result'] = True
            ret['comment'] = comment_port_notexists
        else:
            ret['result'] = None
            ret['comment'] = comment_port_deleted
        return ret

    if bridge and not bridge_exists:
        ret['result'] = False
        ret['comment'] = comment_bridge_notexists
    elif name not in port_list:
        ret['result'] = True
        ret['comment'] = comment_port_notexists
    else:
        if bridge:
            port_remove = __salt__['openvswitch.port_remove'](br=bridge, port=name)
        else:
            port_remove = __salt__['openvswitch.port_remove'](br=None, port=name)

        if port_remove:
            ret['result'] = True
            ret['comment'] = comment_port_deleted
            ret['changes'] = changes_port_deleted
        else:
            ret['result'] = False
            ret['comment'] = comment_port_notdeleted

    return ret



# -*- coding: utf-8 -*-
'''
Manage groups on Solaris

.. important::
    If you feel that Salt should be using this module to manage groups on a
    minion, and it is using a different module (or gives an error similar to
    *'group.info' is not available*), see :ref:`here
    <module-provider-override>`.
'''
from __future__ import absolute_import

# Import python libs
import logging

# Import salt libs
import salt.utils


log = logging.getLogger(__name__)


try:
    import grp
except ImportError:
    pass

# Define the module's virtual name
__virtualname__ = 'group'


def __virtual__():
    '''
    Set the group module if the kernel is SunOS
    '''
    if __grains__['kernel'] == 'SunOS':
        return __virtualname__
    return (False, 'The solaris_group execution module failed to load: '
            'only available on Solaris systems.')


def add(name, gid=None, **kwargs):
    '''
    Add the specified group

    CLI Example:

    .. code-block:: bash

        salt '*' group.add foo 3456
    '''
    if salt.utils.is_true(kwargs.pop('system', False)):
        log.warning('solaris_group module does not support the \'system\' '
                    'argument')
    if kwargs:
        log.warning('Invalid kwargs passed to group.add')

    cmd = 'groupadd '
    if gid:
        cmd += '-g {0} '.format(gid)
    cmd += name

    ret = __salt__['cmd.run_all'](cmd, python_shell=False)

    return not ret['retcode']


def delete(name):
    '''
    Remove the named group

    CLI Example:

    .. code-block:: bash

        salt '*' group.delete foo
    '''
    ret = __salt__['cmd.run_all']('groupdel {0}'.format(name), python_shell=False)

    return not ret['retcode']


def info(name):
    '''
    Return information about a group

    CLI Example:

    .. code-block:: bash

        salt '*' group.info foo
    '''
    try:
        grinfo = grp.getgrnam(name)
    except KeyError:
        return {}
    else:
        return {'name': grinfo.gr_name,
                'passwd': grinfo.gr_passwd,
                'gid': grinfo.gr_gid,
                'members': grinfo.gr_mem}


def getent(refresh=False):
    '''
    Return info on all groups

    CLI Example:

    .. code-block:: bash

        salt '*' group.getent
    '''
    if 'group.getent' in __context__ and not refresh:
        return __context__['group.getent']

    ret = []
    for grinfo in grp.getgrall():
        ret.append(info(grinfo.gr_name))

    __context__['group.getent'] = ret
    return ret


def chgid(name, gid):
    '''
    Change the gid for a named group

    CLI Example:

    .. code-block:: bash

        salt '*' group.chgid foo 4376
    '''
    pre_gid = __salt__['file.group_to_gid'](name)
    if gid == pre_gid:
        return True
    cmd = 'groupmod -g {0} {1}'.format(gid, name)
    __salt__['cmd.run'](cmd, python_shell=False)
    post_gid = __salt__['file.group_to_gid'](name)
    if post_gid != pre_gid:
        return post_gid == gid
    return False



# -*- coding: utf-8 -*-
'''
The cp module is used to execute the logic used by the salt-cp command
line application, salt-cp is NOT intended to broadcast large files, it is
intended to handle text files.
Salt-cp can be used to distribute configuration files
'''

# Import python libs
from __future__ import print_function
from __future__ import absolute_import
import os
import sys

# Import salt libs
import salt.client
from salt.utils import parsers, print_cli
from salt.utils.verify import verify_log
import salt.output


class SaltCPCli(parsers.SaltCPOptionParser):
    '''
    Run the salt-cp command line client
    '''

    def run(self):
        '''
        Execute salt-cp
        '''
        self.parse_args()

        # Setup file logging!
        self.setup_logfile_logger()
        verify_log(self.config)

        cp_ = SaltCP(self.config)
        cp_.run()


class SaltCP(object):
    '''
    Create a salt cp object, used to distribute simple files with salt
    '''
    def __init__(self, opts):
        self.opts = opts

    def _file_dict(self, fn_):
        '''
        Take a path and return the contents of the file as a string
        '''
        if not os.path.isfile(fn_):
            err = 'The referenced file, {0} is not available.'.format(fn_)
            sys.stderr.write(err + '\n')
            sys.exit(42)
        with salt.utils.fopen(fn_, 'r') as fp_:
            data = fp_.read()
        return {fn_: data}

    def _recurse_dir(self, fn_, files=None):
        '''
        Recursively pull files from a directory
        '''
        if files is None:
            files = {}

        for base in os.listdir(fn_):
            path = os.path.join(fn_, base)
            if os.path.isdir(path):
                files.update(self._recurse_dir(path))
            else:
                files.update(self._file_dict(path))
        return files

    def _load_files(self):
        '''
        Parse the files indicated in opts['src'] and load them into a python
        object for transport
        '''
        files = {}
        for fn_ in self.opts['src']:
            if os.path.isfile(fn_):
                files.update(self._file_dict(fn_))
            elif os.path.isdir(fn_):
                print_cli(fn_ + ' is a directory, only files are supported.')
                #files.update(self._recurse_dir(fn_))
        return files

    def run(self):
        '''
        Make the salt client call
        '''
        arg = [self._load_files(), self.opts['dest']]
        local = salt.client.get_local_client(self.opts['conf_file'])
        args = [self.opts['tgt'],
                'cp.recv',
                arg,
                self.opts['timeout'],
                ]

        selected_target_option = self.opts.get('selected_target_option', None)
        if selected_target_option is not None:
            args.append(selected_target_option)

        ret = local.cmd(*args)

        salt.output.display_output(
                ret,
                self.opts.get('output', 'nested'),
                self.opts)



# -*- coding: utf-8 -*-
'''
    :codeauthor: :email:`Jayesh Kariya <jayeshk@saltstack.com>`
'''
# Import Python libs
from __future__ import absolute_import

# Import Salt Testing Libs
from salttesting import skipIf, TestCase
from salttesting.mock import (
    NO_MOCK,
    NO_MOCK_REASON,
    MagicMock,
    call,
    patch)

from salttesting.helpers import ensure_in_syspath

ensure_in_syspath('../../')

# Import Salt Libs
from salt.states import ipset

ipset.__salt__ = {}
ipset.__opts__ = {}


@skipIf(NO_MOCK, NO_MOCK_REASON)
class IpsetSetPresentTestCase(TestCase):
    '''
    Test cases for salt.states.ipset.present
    '''

    fake_name = 'fake_ipset'
    fake_set_type = {'bitmap': '192.168.0.3'}

    def _runner(self, expected_ret, test=False, check_set=False, new_set=None,
                new_set_assertion=True):
        mock_check_set = MagicMock(return_value=check_set)
        mock_new_set = MagicMock() if new_set is None else MagicMock(return_value=new_set)
        with patch.dict(ipset.__salt__, {'ipset.check_set': mock_check_set,
                                         'ipset.new_set': mock_new_set}):
            with patch.dict(ipset.__opts__, {'test': test}):
                actual_ret = ipset.set_present(self.fake_name, self.fake_set_type)
        mock_check_set.assert_called_once_with(self.fake_name)
        if new_set_assertion:
            mock_new_set.assert_called_once_with(self.fake_name, self.fake_set_type, 'ipv4')
        else:
            mock_new_set.assert_not_called()
        self.assertDictEqual(actual_ret, expected_ret)

    def test_already_exists(self):
        '''
        Test to verify the chain exists when it already exists.
        '''
        ret = {'name': self.fake_name,
               'result': True,
               'comment': 'ipset set {0} already exists for ipv4'.format(self.fake_name),
               'changes': {}}
        self._runner(ret, check_set=True, new_set_assertion=False)

    def test_needs_update_test_mode(self):
        '''
        Test to verify that detects need for update but doesn't apply when in test mode.
        '''

        ret = {'name': self.fake_name,
               'result': None,
               'comment': 'ipset set {0} would be added for ipv4'.format(self.fake_name),
               'changes': {}}
        self._runner(ret, test=True, new_set_assertion=False)

    def test_creates_set(self):
        ret = {'name': self.fake_name,
               'result': True,
               'comment': 'ipset set {0} created successfully for ipv4'.format(self.fake_name),
               'changes': {'locale': self.fake_name}}
        self._runner(ret, new_set=True)

    def test_create_fails(self):
        ret = {'name': self.fake_name,
               'result': False,
               'comment': 'Failed to create set {0} for ipv4: '.format(self.fake_name),
               'changes': {}}
        self._runner(ret, new_set='')


@skipIf(NO_MOCK, NO_MOCK_REASON)
class IpsetSetAbsentTestCase(TestCase):
    '''
    Test cases for salt.states.ipset.present
    '''

    fake_name = 'fake_ipset'
    fake_set_type = {'bitmap': '192.168.0.3'}

    def _runner(self, expected_ret, test=False, check_set=True, delete_set='',
                flush_assertion=False, delete_set_assertion=False):
        mock_check_set = MagicMock(return_value=check_set)
        mock_flush = MagicMock()
        mock_delete_set = MagicMock() if delete_set is None else MagicMock(return_value=delete_set)
        with patch.dict(ipset.__opts__, {'test': test}):
            with patch.dict(ipset.__salt__, {'ipset.check_set': mock_check_set,
                                             'ipset.flush': mock_flush,
                                             'ipset.delete_set': mock_delete_set}):
                actual_ret = ipset.set_absent(self.fake_name)
        mock_check_set.assert_called_once_with(self.fake_name, 'ipv4')
        if flush_assertion:
            mock_flush.assert_called_once_with(self.fake_name, 'ipv4')
        else:
            mock_flush.assert_not_called()
        if delete_set_assertion:
            mock_delete_set.assert_called_once_with(self.fake_name, 'ipv4')
        else:
            mock_delete_set.assert_not_called()
        self.assertDictEqual(actual_ret, expected_ret)

    def test_already_absent(self):
        ret = {'name': self.fake_name,
               'result': True,
               'comment': 'ipset set {0} for ipv4 is already absent'.format(self.fake_name),
               'changes': {}}
        self._runner(ret, check_set=False, delete_set=None)

    def test_remove_test_mode(self):
        ret = {'name': self.fake_name,
               'result': None,
               'comment': 'ipset set {0} for ipv4 would be removed'.format(self.fake_name),
               'changes': {}}
        self._runner(ret, test=True, delete_set=None)

    def test_remove_fails(self):
        ret = {'name': self.fake_name,
               'result': False,
               'comment': 'Failed to delete set {0} for ipv4: '.format(self.fake_name),
               'changes': {}}
        self._runner(ret, flush_assertion=True, delete_set_assertion=True)

    def test_remove_success(self):
        ret = {'name': self.fake_name,
               'result': True,
               'comment': 'ipset set {0} deleted successfully for family ipv4'.format(self.fake_name),
               'changes': {'locale': 'fake_ipset'}}
        self._runner(ret, delete_set=True, flush_assertion=True, delete_set_assertion=True)


@skipIf(NO_MOCK, NO_MOCK_REASON)
class IpsetPresentTestCase(TestCase):
    '''
    Test cases for salt.states.ipset.present
    '''

    fake_name = 'fake_ipset'
    fake_entries = ['192.168.0.3', '192.168.1.3']

    def _runner(self, expected_ret, test=False, check=False, add=False,
                add_assertion=False):
        mock_check = MagicMock(return_value=check)
        mock_add = MagicMock(return_value=add)
        with patch.dict(ipset.__opts__, {'test': test}):
            with patch.dict(ipset.__salt__, {'ipset.check': mock_check,
                                             'ipset.add': mock_add}):
                actual_ret = ipset.present(self.fake_name, self.fake_entries, set_name=self.fake_name)

        mock_check.assert_has_calls([call(self.fake_name, e, 'ipv4') for e in self.fake_entries], any_order=True)
        if add_assertion:
            expected_calls = [call(self.fake_name, e, 'ipv4', set_name=self.fake_name) for e in self.fake_entries]
            if add is not True:
                # if the add fails, then it will only get called once.
                expected_calls = expected_calls[:1]
            mock_add.assert_has_calls(expected_calls, any_order=True)
        else:
            mock_add.assert_not_called()
        self.assertDictEqual(actual_ret, expected_ret)

    def test_entries_already_present(self):
        ret = {'name': self.fake_name,
               'result': True,
               'comment': 'entry for 192.168.0.3 already in set {0} for ipv4\n'
                          'entry for 192.168.1.3 already in set {0} for ipv4\n'
                          ''.format(self.fake_name),
               'changes': {}}
        self._runner(ret, check=True)

    def test_in_test_mode(self):
        ret = {'name': self.fake_name,
               'result': None,
               'comment': 'entry 192.168.0.3 would be added to set {0} for family ipv4\n'
                          'entry 192.168.1.3 would be added to set {0} for family ipv4\n'
                          ''.format(self.fake_name),
               'changes': {}}
        self._runner(ret, test=True)

    def test_add_fails(self):
        ret = {'name': self.fake_name,
               'result': False,
               'comment': 'Failed to add to entry 192.168.1.3 to set {0} for family ipv4.\n'
                          'Error'.format(self.fake_name),
               'changes': {}}
        self._runner(ret, add='Error', add_assertion=True)

    def test_success(self):
        ret = {'name': self.fake_name,
               'result': True,
               'comment': 'entry 192.168.0.3 added to set {0} for family ipv4\n'
                          'entry 192.168.1.3 added to set {0} for family ipv4\n'
                          ''.format(self.fake_name),
               'changes': {'locale': 'fake_ipset'}}
        self._runner(ret, add='worked', add_assertion=True)

    def test_missing_entry(self):
        ret = {'name': self.fake_name,
               'result': False,
               'comment': 'ipset entry must be specified',
               'changes': {}}
        mock = MagicMock(return_value=True)
        with patch.dict(ipset.__salt__, {'ipset.check': mock}):
            self.assertDictEqual(ipset.present(self.fake_name), ret)


@skipIf(NO_MOCK, NO_MOCK_REASON)
class IpsetAbsentTestCase(TestCase):
    '''
    Test cases for salt.states.ipset.present
    '''

    fake_name = 'fake_ipset'
    fake_entries = ['192.168.0.3', '192.168.1.3']

    def _runner(self, expected_ret, test=False, check=False, delete=False,
                delete_assertion=False):
        mock_check = MagicMock(return_value=check)
        mock_delete = MagicMock(return_value=delete)
        with patch.dict(ipset.__opts__, {'test': test}):
            with patch.dict(ipset.__salt__, {'ipset.check': mock_check,
                                             'ipset.delete': mock_delete}):
                actual_ret = ipset.absent(self.fake_name, self.fake_entries, set_name=self.fake_name)
        mock_check.assert_has_calls([call(self.fake_name, e, 'ipv4') for e in self.fake_entries], any_order=True)
        if delete_assertion:
            expected_calls = [call(self.fake_name, e, 'ipv4', set_name=self.fake_name) for e in self.fake_entries]
            if delete is not True:
                expected_calls = expected_calls[:1]
            mock_delete.assert_has_calls(expected_calls, any_order=True)
        else:
            mock_delete.assert_not_called()
        self.assertDictEqual(actual_ret, expected_ret)

    def test_already_absent(self):
        ret = {'name': self.fake_name,
               'result': True,
               'comment': 'ipset entry for 192.168.0.3 not present in set {0} for ipv4\n'
                          'ipset entry for 192.168.1.3 not present in set {0} for ipv4\n'
                          ''.format(self.fake_name),
               'changes': {}}
        self._runner(ret)

    def test_in_test_mode(self):
        ret = {'name': self.fake_name,
               'result': None,
               'comment': 'ipset entry 192.168.0.3 would be removed from set {0} for ipv4\n'
                          'ipset entry 192.168.1.3 would be removed from set {0} for ipv4\n'
                          ''.format(self.fake_name),
               'changes': {}}
        self._runner(ret, test=True, check=True)

    def test_del_fails(self):
        ret = {'name': self.fake_name,
               'result': False,
               'comment': 'Failed to delete ipset entry from set {0} for ipv4. Attempted entry was 192.168.1.3.\n'
                          'Error\n'.format(self.fake_name),
               'changes': {}}
        self._runner(ret, check=True, delete='Error', delete_assertion=True)

    def test_success(self):
        ret = {'name': self.fake_name,
               'result': True,
               'comment': 'ipset entry 192.168.0.3 removed from set {0} for ipv4\n'
                          'ipset entry 192.168.1.3 removed from set {0} for ipv4\n'
                          ''.format(self.fake_name),
               'changes': {'locale': 'fake_ipset'}}
        self._runner(ret, check=True, delete='worked', delete_assertion=True)

    def test_absent(self):
        ret = {'name': self.fake_name,
               'result': False,
               'comment': 'ipset entry must be specified',
               'changes': {}}
        mock = MagicMock(return_value=True)
        with patch.dict(ipset.__salt__, {'ipset.check': mock}):
            self.assertDictEqual(ipset.absent(self.fake_name), ret)


@skipIf(NO_MOCK, NO_MOCK_REASON)
class IpsetFlushTestCase(TestCase):
    '''
    Test cases for salt.states.ipset.present
    '''

    fake_name = 'fake_ipset'

    def _runner(self, expected_ret, test=False, check_set=True, flush=True,
                flush_assertion=True):
        mock_check_set = MagicMock(return_value=check_set)
        mock_flush = MagicMock(return_value=flush)
        with patch.dict(ipset.__opts__, {'test': test}):
            with patch.dict(ipset.__salt__, {'ipset.check_set': mock_check_set,
                                             'ipset.flush': mock_flush}):
                actual_ret = ipset.flush(self.fake_name)
        mock_check_set.assert_called_once_with(self.fake_name)
        if flush_assertion:
            mock_flush.assert_called_once_with(self.fake_name, 'ipv4')
        else:
            mock_flush.assert_not_called()
        self.assertDictEqual(actual_ret, expected_ret)

    def test_no_set(self):
        ret = {'name': self.fake_name,
               'result': False,
               'comment': 'ipset set {0} does not exist for ipv4'.format(self.fake_name),
               'changes': {}}
        self._runner(ret, check_set=False, flush_assertion=False)

    def test_in_test_mode(self):
        ret = {'name': self.fake_name,
               'result': None,
               'comment': 'ipset entries in set {0} for ipv4 would be flushed'.format(self.fake_name),
               'changes': {}}
        self._runner(ret, test=True, flush_assertion=False)

    def test_flush_fails(self):
        ret = {'name': self.fake_name,
               'result': False,
               'comment': 'Failed to flush ipset entries from set {0} for ipv4'.format(self.fake_name),
               'changes': {}}
        self._runner(ret, flush=False)

    def test_success(self):
        ret = {'name': self.fake_name,
               'result': True,
               'comment': 'Flushed ipset entries from set {0} for ipv4'.format(self.fake_name),
               'changes': {'locale': 'fake_ipset'}}
        self._runner(ret)



# -*- coding: utf-8 -*-

# Import Python Libs
from __future__ import absolute_import
import os
import textwrap

# Import Salt Testing libs
from salttesting.helpers import ensure_in_syspath

ensure_in_syspath('../../')

# Import Salt Libs
import integration
import salt.utils


STATE_DIR = os.path.join(integration.FILES, 'file', 'base')


class EnabledTest(integration.ModuleCase):
    '''
    validate the use of shell processing for cmd.run on the salt command line
    and in templating
    '''
    cmd = ("printf '%s\n' first second third | wc -l ; "
           "export SALTY_VARIABLE='saltines' && echo $SALTY_VARIABLE ; "
           "echo duh &> /dev/null")

    def test_shell_default_enabled(self):
        '''
        ensure that python_shell defaults to True for cmd.run
        '''
        enabled_ret = '3\nsaltines'  # the result of running self.cmd in a shell
        ret = self.run_function('cmd.run', [self.cmd])
        self.assertEqual(ret.strip(), enabled_ret)

    def test_shell_disabled(self):
        '''
        test shell disabled output for cmd.run
        '''
        disabled_ret = ('first\nsecond\nthird\n|\nwc\n-l\n;\nexport\nSALTY_VARIABLE=saltines'
                        '\n&&\necho\n$SALTY_VARIABLE\n;\necho\nduh\n&>\n/dev/null')
        ret = self.run_function('cmd.run', [self.cmd], python_shell=False)
        self.assertEqual(ret, disabled_ret)

    def test_template_default_enabled(self):
        '''
        ensure that python_shell defaults to True for templates
        '''
        state_name = 'template_shell_enabled'
        state_filename = state_name + '.sls'
        state_file = os.path.join(STATE_DIR, state_filename)

        enabled_ret = '3 saltines'  # the result of running self.cmd in a shell
        ret_key = 'test_|-shell_enabled_|-{0}_|-configurable_test_state'.format(enabled_ret)

        try:
            with salt.utils.fopen(state_file, 'w') as fp_:
                fp_.write(textwrap.dedent('''\
                {{% set shell_enabled = salt['cmd.run']("{0}").strip() %}}

                shell_enabled:
                  test.configurable_test_state:
                    - name: '{{{{ shell_enabled }}}}'
                '''.format(self.cmd)))

            ret = self.run_function('state.sls', [state_name])
            self.assertEqual(ret[ret_key]['name'], enabled_ret)
        finally:
            os.remove(state_file)

    def test_template_disabled(self):
        '''
        test shell disabled output for templates
        '''
        state_name = 'template_shell_disabled'
        state_filename = state_name + '.sls'
        state_file = os.path.join(STATE_DIR, state_filename)

        # the result of running self.cmd not in a shell
        disabled_ret = ('first second third | wc -l ; export SALTY_VARIABLE=saltines '
                        '&& echo $SALTY_VARIABLE ; echo duh &> /dev/null')
        ret_key = 'test_|-shell_enabled_|-{0}_|-configurable_test_state'.format(disabled_ret)

        try:
            with salt.utils.fopen(state_file, 'w') as fp_:
                fp_.write(textwrap.dedent('''\
                {{% set shell_disabled = salt['cmd.run']("{0}", python_shell=False) %}}

                shell_enabled:
                  test.configurable_test_state:
                    - name: '{{{{ shell_disabled }}}}'
                '''.format(self.cmd)))

            ret = self.run_function('state.sls', [state_name])
            self.assertEqual(ret[ret_key]['name'], disabled_ret)
        finally:
            os.remove(state_file)


if __name__ == '__main__':
    from integration import run_tests
    run_tests(EnabledTest)



# -*- coding: utf-8 -*-
'''
Detect MDADM RAIDs
'''
from __future__ import absolute_import

# Import python libs
import logging

# Import salt libs
import salt.utils

log = logging.getLogger(__name__)


def mdadm():
    '''
    Return list of mdadm devices
    '''
    devices = set()
    try:
        with salt.utils.fopen('/proc/mdstat', 'r') as mdstat:
            for line in mdstat:
                if line.startswith('Personalities : '):
                    continue
                if line.startswith('unused devices:'):
                    continue
                if ' : ' in line:
                    devices.add(line.split(' : ')[0])
    except IOError:
        return {}

    devices = sorted(devices)
    if devices:
        log.trace(
            'mdadm devices detected: {0}'.format(', '.join(devices))
        )

    return {'mdadm': devices}



# -*- coding: utf-8 -*-
'''
 Calculate the difference between two dictionaries as:
    (1) items added
    (2) items removed
    (3) keys same in both but changed values
    (4) keys same in both and unchanged values

  Originally posted at http://stackoverflow.com/questions/1165352/fast-comparison-between-two-python-dictionary/1165552#1165552
  Available at repository: https://github.com/hughdbrown/dictdiffer
'''


def diff(current_dict, past_dict):
    return DictDiffer(current_dict, past_dict)


class DictDiffer(object):
    """
    Calculate the difference between two dictionaries as:
    (1) items added
    (2) items removed
    (3) keys same in both but changed values
    (4) keys same in both and unchanged values


    """
    def __init__(self, current_dict, past_dict):
        self.current_dict, self.past_dict = current_dict, past_dict
        self.set_current, self.set_past = set(current_dict.keys()), set(past_dict.keys())
        self.intersect = self.set_current.intersection(self.set_past)

    def added(self):
        return self.set_current - self.intersect

    def removed(self):
        return self.set_past - self.intersect

    def changed(self):
        return set(o for o in self.intersect if self.past_dict[o] != self.current_dict[o])

    def unchanged(self):
        return set(o for o in self.intersect if self.past_dict[o] == self.current_dict[o])
